import os
import ctypes as ct
import numpy as np
import pandas as pd
import time

from .enums import E_MODAL_T, E_PUPIL_DETECTION_T, E_ZERNIKE_NORM_T, E_PUPIL_COVERING_T

basepath = "D:\\WaveKitX64"

imop_lib = ct.windll.LoadLibrary(
    os.path.join(basepath, "C", "Lib", "c_interface_vc100_x64.dll")
)

mode_names = [
    "Vert. Tilt",
    "Horz. Tilt",
    "Defocus",
    "Vert. Asm.",
    "Oblq. Asm.",
    "Vert. Coma",
    "Horz. Coma",
    "3rd Spherical",
    "Vert. Tre.",
    "Horz. Tre.",
    "Vert. 5th Asm.",
    "Oblq. 5th Asm.",
    "Vert. 5th Coma",
    "Horz. 5th Coma",
    "5th Spherical",
    "Vert. Tetra.",
    "Oblq. Tetra.",
    "Vert. 7th Tre.",
    "Horz. 7th Tre.",
    "Vert. 7th Asm.",
    "Oblq. 7th Asm.",
    "Vert. 7th Coma",
    "Horz. 7th Coma",
    "7th Spherical",
    "Vert. Penta.",
    "Horz. Penta.",
    "Vert. 9th Tetra.",
    "Oblq. 9th Tetra.",
    "Vert. 9th Tre.",
    "Horz. 9th Tre.",
    "Vert. 9th Asm.",
    "Oblq. 9th Asm.",
]

# ctypes utility functions:
def char_p(s):
    # return a C character pointer from python String
    return ct.c_char_p(s.encode("utf-8"))


def arr_p(ndarr):
    # convert numpy.ndarray to C pointer array
    return ct.c_void_p(ndarr.ctypes.data)


def call_errmsg(func, *args):
    message = ct.create_string_buffer(256)

    flag = func(message, *args)

    err_msg = f"{func.__name__}: {message.value.decode('utf-8')}"
    del message

    if flag:
        raise Exception(err_msg)

    return err_msg, flag


# CTypes Structures:


class uint2D(ct.Structure):

    _fields_ = [("X", ct.c_uint), ("Y", ct.c_uint)]


class int2D(ct.Structure):

    _fields_ = [("X", ct.c_int), ("Y", ct.c_int)]


class float2D(ct.Structure):

    _fields_ = [("X", ct.c_float), ("Y", ct.c_float)]


# Generic class which contains a pointer to object address
# (used in C Imop functions)
class Pointer:
    def __init__(self):
        self.pointer = ct.c_void_p()

    def __call__(self):
        return self.pointer


class IMOP_Mirror:
    def __init__(
        self,
        wfc_config_file_path=os.path.join(
            basepath, "MirrorFiles", "WaveFrontCorrector_Mirao52-e_0259.dat"
        ),
        haso_config_file_path=os.path.join(
            basepath, "MirrorFiles", "HASO4_first_7458.dat"
        ),
        positions_file_path=os.path.join(
            basepath, "MirrorFiles", "OPMv3_SysCorr_517nm_20230324.wcs"
        ),
        interaction_matrix_file_path=os.path.join(
            basepath, "MirrorFiles", "VAST_Sept_2023_b.aoc"
        ),
        n_modes=32,
    ):

        mirror_set = WavefrontCorrectorSet()
        mirror_set.new_from_config_file(wfc_config_file_path)

        mirror = WavefrontCorrector(
            wfc_config_file_path=wfc_config_file_path,
            haso_config_file_path=haso_config_file_path,
            positions_file_path=positions_file_path,
            interaction_matrix_file_path=interaction_matrix_file_path,
            n_actuators=mirror_set.get_actuators_count(),
        )

        corrdata_manager = CorrDataManager()
        corrdata_manager.new_from_backup_file(
            haso_config_file_path, interaction_matrix_file_path
        )

        corrdata_manager.set_command_matrix_prefs(nb_kept_modes=32)
        corrdata_manager.compute_command_matrix()

        mirror.set_temporization(20)

        config = CoreEngine().get_config(haso_config_file_path)

        pupil = Pupil()
        pupil.new_from_dimensions(config.nb_subpupils, config.ulens_step, 0)

        pupil_zeros = np.zeros((config.nb_subpupils.Y, config.nb_subpupils.X))
        pupil_buffer = corrdata_manager.get_greatest_common_pupil(pupil_zeros)

        pupil.set_data(pupil_buffer)

        self.pupil_center, self.pupil_radius = PupilCompute().fit_zernike_pupil(
            pupil,
            E_PUPIL_DETECTION_T.E_PUPIL_AUTOMATIC,
            E_PUPIL_COVERING_T.E_PUPIL_CIRCUMSCRIBED,
            0,
        )

        self.position_flat = mirror.get_positions_from_file(positions_file_path)

        modal_coef = ModalCoef()
        modal_coef.new()  # default is Zernike
        modal_coef.set_zernike_prefs(
            n_modes,
            0,
            np.empty(0),
            self.pupil_center,
            self.pupil_radius
            # normalization=E_ZERNIKE_NORM_T.E_ZERNIKE_NORM_RMS
        )

        self.n_modes = n_modes
        self.mirror = mirror
        self.mirror_set = mirror_set
        self.corrdata_manager = corrdata_manager
        self.pupil = pupil
        self.modal_coef = modal_coef
        self.haso_slopes = HasoSlopes()
        self.last_delta_commands = np.zeros(self.mirror.n_actuators, dtype=np.float32)

    def flat(self):
        self.display_modes(np.zeros(self.n_modes, dtype=np.float32))

    def zero_flatness(self):
        self.set_flat(np.zeros(self.mirror.n_actuators, dtype=np.float32))
        self.flat()

    def set_flat(self, pos=None, pos_path=None):
        if pos_path:
            pos = self.mirror.get_positions_from_file(pos_path)
        self.position_flat = pos

    def move_absolute_zero(self):
        self.mirror.move_absolute(np.zeros(self.mirror.n_actuators, dtype=np.float32))

    def display_modes(self, coefs, wait=False):
        # make sure coefs is np.float32 array
        if type(coefs) is not np.ndarray:
            coefs = np.array(coefs)
        if coefs.dtype is not np.float32:
            coefs = coefs.astype(np.float32)

        try:
            self.modal_coef.set_data(
                coefs,
                np.arange(1, self.n_modes + 1).astype(np.uint32),
                self.n_modes,
                self.pupil,
            )
        except Exception as modal_coef_exception:
            print("modal_coef.set_data failed...\n", modal_coef_exception)

        try:
            self.haso_slopes.new_from_modal_coef(
                self.modal_coef, self.mirror.haso_config_file_path
            )
        except Exception as haso_exception:
            print("haso_slopes.new_from_modal_coef failed...\n", haso_exception)

        try:
            delta_commands = (
                self.corrdata_manager.compute_delta_command_from_delta_slopes(
                    self.haso_slopes,
                    np.zeros(self.mirror.n_actuators).astype(np.float32),
                )
            )
        except Exception as delta_commands_exception:
            print(
                "corrdata_manager.compute_delta_command_from_delta_slopes failed...\n",
                delta_commands_exception,
            )

        self.last_delta_commands = delta_commands

        new_positions = self.position_flat + delta_commands

        try:
            self.mirror.move_absolute(new_positions)
        except Exception as move_exception:
            print("mirror.move_absolute failed...\n", move_exception)

        if wait:
            # TODO: doesn't work very well...
            timeout = 1000
            t = 0
            while (self.get_modal_coefs()[0] != coefs).any():
                t += 1
                if t == timeout:
                    print("Ran out of time...")
                    break
                time.sleep(0.001)

    def update_delta_commands(self):
        self.last_delta_commands = (
            self.mirror.get_current_positions() - self.position_flat
        )

    def get_modal_coefs(self):
        return self.modal_coef.get_data(n_modes=self.n_modes, pupil=self.pupil)

    def load_wcs(self, path=None, name=None, mode_file=False):
        if path:
            wcs_load_path = path
        elif name:
            wcs_load_path = os.path.join(basepath, "MirrorFiles", name + ".wcs")
        else:
            print("IMOP_Mirror:: Need to provide either name or path!")
            return

        new_positions = self.mirror.get_positions_from_file(wcs_load_path)
        self.mirror.move_absolute(new_positions)
        self.update_delta_commands()

        if mode_file:
            mode_load_path = wcs_load_path.split(".")[0] + ".json"

            try:
                import json

                with open(mode_load_path, "r") as f:
                    coefs = json.load(f)
            except FileNotFoundError as err:
                print(f"{err.strerror}::{mode_load_path} -> Modal coef not updated...")
                return []

            coefs = [float(c) for c in coefs.values()]

            return coefs

    def save_wcs(self, path=None, name=None, mode_file=False):
        if path:
            wcs_save_path = path
        elif name:
            wcs_save_path = os.path.join(basepath, "MirrorFiles", name + ".wcs")
        else:
            print("IMOP_Mirror:: Need to provide either name or path!")
            return

        self.mirror.save_positions_to_file(file_path=wcs_save_path)

        if mode_file:
            mode_save_path = wcs_save_path.split(".")[0] + ".json"
            coefs, coef_inds = self.get_modal_coefs()
            mode_dict = {}
            for c in coef_inds:
                mode_dict[mode_names[c - 1]] = f"{coefs[c-1]:.4f}"

            import json

            with open(mode_save_path, "w") as f:
                json.dump(mode_dict, f)
            f.close()

    def get_wavefront_pix(self):
        radius = 4
        x, y = np.meshgrid(range(2 * radius), range(2 * radius))

        x = x - x.mean()
        y = y - y.mean()

        pupil = x**2 + y**2 <= radius**2
        wavefront = np.zeros(pupil.shape)

        wavefront[pupil[:]] = self.last_delta_commands

        return wavefront


# CoreEngine class (cHasoConfig.h)
class CoreEngine:
    def get_config(self, haso_config_file_path):
        _, _, config = self.__Imop_CoreEngine_GetConfig(haso_config_file_path)

        return pd.Series(config)

    def __Imop_CoreEngine_GetConfig(self, haso_config_file_path):
        serial_number = ct.create_string_buffer(256)
        revision = ct.c_uint()
        model = ct.create_string_buffer(256)
        nb_subpupils = uint2D()
        ulens_step = float2D()
        alignment_position_pixels = float2D()
        tolerance_radius = ct.c_ushort()
        default_start_subpupil = uint2D()
        lower_calibration_wavelen = ct.c_double()
        upper_calibration_wavelen = ct.c_double()
        black_subpupil_position = uint2D()
        tilt_limit_mrad = ct.c_double()
        radius = ct.c_double()
        microlens_focal = ct.c_double()
        smearing_limit_wavelen = ct.c_double()
        smearing_limit_exp_duration = ct.c_double()
        internal_options_list = ct.create_string_buffer(256)
        software_info_list = ct.create_string_buffer(256)
        SdkInfoList = ct.create_string_buffer(256)

        msg, result = call_errmsg(
            imop_lib.Imop_CoreEngine_GetConfig,
            char_p(haso_config_file_path),
            serial_number,
            ct.byref(revision),
            model,
            ct.byref(nb_subpupils),
            ct.byref(ulens_step),
            ct.byref(alignment_position_pixels),
            ct.byref(tolerance_radius),
            ct.byref(default_start_subpupil),
            ct.byref(lower_calibration_wavelen),
            ct.byref(upper_calibration_wavelen),
            ct.byref(black_subpupil_position),
            ct.byref(tilt_limit_mrad),
            ct.byref(radius),
            ct.byref(microlens_focal),
            ct.byref(smearing_limit_wavelen),
            ct.byref(smearing_limit_exp_duration),
            internal_options_list,
            software_info_list,
            SdkInfoList,
        )

        config = {
            "serial_number": serial_number.value.decode("utf-8"),
            "revision": revision.value,
            "model": model.value.decode("utf-8"),
            "nb_subpupils": nb_subpupils,
            "ulens_step": ulens_step,
            "alignment_position_pixels": alignment_position_pixels,
            "tolerance_radius": tolerance_radius.value,
            "default_start_subpupil": default_start_subpupil,
            "lower_calibration_wavelen": lower_calibration_wavelen.value,
            "upper_calibration_wavelen": upper_calibration_wavelen.value,
            "black_subpupil_position": black_subpupil_position,
            "tilt_limit_mrad": tilt_limit_mrad.value,
            "radius": radius.value,
            "microlens_focal": microlens_focal.value,
            "smearing_limit_wavelen": smearing_limit_wavelen.value,
            "smearing_limit_exp_duration": smearing_limit_exp_duration.value,
            "internal_options_list": internal_options_list.value.decode("utf-8").split(
                ";"
            ),
            "software_info_list": software_info_list.value.decode("utf-8").split(";"),
            "SdkInfoList": SdkInfoList.value.decode("utf-8").split(";"),
        }

        return msg, result, config


# HasoSlopes class (cHasoSlopes.h)
class HasoSlopes(Pointer):
    def __init__(self):
        super().__init__()

    # Public functions:

    def new_from_modal_coef(self, modal_coef, haso_config_file_path):
        self.__Imop_HasoSlopes_NewFromModalCoef(modal_coef, haso_config_file_path)

    # Private functions translated directly from Imop C API:

    def __Imop_HasoSlopes_NewFromModalCoef(self, modal_coef, config_file_path):
        return call_errmsg(
            imop_lib.Imop_HasoSlopes_NewFromModalCoef,
            ct.byref(self.pointer),
            modal_coef.pointer,
            char_p(config_file_path),
        )


# PupilCompute class (cComputePupil.h)
class PupilCompute(Pointer):
    def __init__(self):
        super().__init__()

    # Public functions:

    def fit_zernike_pupil(
        self,
        pupil,
        detection_mode=E_PUPIL_DETECTION_T.E_PUPIL_AUTOMATIC,
        covering=E_PUPIL_COVERING_T.E_PUPIL_INSCRIBED,
        has_central_occultation=False,
    ):
        _, _, center, radius = self.__Imop_PupilCompute_FitZernikePupil(
            pupil, detection_mode, covering, has_central_occultation
        )

        return center, radius

    # Private functions translated directly from Imop C API:

    def __Imop_PupilCompute_FitZernikePupil(
        self, pupil, detection_mode, covering, has_central_occultation
    ):
        center = float2D()
        radius = ct.c_float()

        msg, result = call_errmsg(
            imop_lib.Imop_PupilCompute_FitZernikePupil,
            pupil.pointer,
            detection_mode,
            covering,
            has_central_occultation,
            ct.byref(center),
            ct.byref(radius),
        )

        return msg, result, center, radius


# Pupil class (cPupil.h)
class Pupil(Pointer):
    def __init__(self):
        super().__init__()

    def __del__(self):
        # Destructor
        self.__Imop_Pupil_Delete()

    # Public functions:

    def new_from_dimensions(self, dimensions, steps, value):
        self.__Imop_Pupil_NewFromDimensions(dimensions, steps, value)

    def new_from_zernike_pupil(self, dimensions, steps, center, radius):
        self.__Imop_Pupil_NewFromZernikePupil(dimensions, steps, center, radius)

    def get_data(self, buffer):
        _, _, data = self.__Imop_Pupil_GetData(buffer)

        return data

    def set_data(self, data):
        self.__Imop_Pupil_SetData(data)

    # Private functions translated directly from Imop C API:

    def __Imop_Pupil_NewFromDimensions(self, dimensions, steps, value):
        return call_errmsg(
            imop_lib.Imop_Pupil_NewFromDimensions,
            ct.byref(self.pointer),
            ct.byref(dimensions),
            ct.byref(steps),
            value,
        )

    def __Imop_Pupil_NewFromZernikePupil(self, dimensions, steps, center, radius):
        return call_errmsg(
            imop_lib.Imop_Pupil_NewFromZernikePupil,
            ct.byref(self.pointer),
            ct.byref(steps),
            ct.byref(dimensions),
            ct.byref(center),
            ct.c_float(radius),
        )

    def __Imop_Pupil_GetData(self, data):
        data = data.astype(bool)

        msg, result = call_errmsg(
            imop_lib.Imop_Pupil_GetData, self.pointer, arr_p(data)
        )

        return msg, result, data

    def __Imop_Pupil_SetData(self, data):
        return call_errmsg(imop_lib.Imop_Pupil_SetData, self.pointer, arr_p(data))

    def __Imop_Pupil_Delete(self):
        return call_errmsg(imop_lib.Imop_Pupil_Delete, self.pointer)


# ModalCoef class (cModalCoef.h)
class ModalCoef(Pointer):
    def __init__(self):
        super().__init__()

    def __del__(self):
        # Destructor
        self.__Imop_ModalCoef_Delete()

    # Public functions:

    def new(self, e_modal_t=E_MODAL_T.E_MODALCOEF_ZERNIKE):
        self.__Imop_ModalCoef_New(e_modal_t)

    def set_zernike_prefs(
        self,
        nb_coefs_total,
        nb_coefs_to_filter,
        coefs_to_filter,
        projection_pupil_center,
        projection_pupil_radius,
        normalization=E_ZERNIKE_NORM_T.E_ZERNIKE_NORM_STD,
    ):
        self.__Imop_ModalCoef_SetZernikePrefs(
            normalization,
            nb_coefs_total,
            nb_coefs_to_filter,
            coefs_to_filter,
            projection_pupil_center,
            projection_pupil_radius,
        )

    def set_data(self, coef, index, size, pupil):
        self.__Imop_ModalCoef_SetData(coef, index, size, pupil)

    def get_data(self, n_modes, pupil):
        _, _, coef, index = self.__Imop_ModalCoef_GetData(n_modes, pupil)

        return coef, index

    # Private functions translated directly from Imop C API:

    def __Imop_ModalCoef_New(self, e_modal_t):
        return call_errmsg(
            imop_lib.Imop_ModalCoef_New, ct.byref(self.pointer), e_modal_t
        )

    def __Imop_ModalCoef_SetZernikePrefs(
        self,
        normalization,
        nb_coefs_total,
        nb_coefs_to_filter,
        coefs_to_filter,
        projection_pupil_center,
        projection_pupil_radius,
    ):
        return call_errmsg(
            imop_lib.Imop_ModalCoef_SetZernikePrefs,
            self.pointer,
            normalization,
            nb_coefs_total,
            nb_coefs_to_filter,
            arr_p(coefs_to_filter),
            ct.byref(projection_pupil_center),
            projection_pupil_radius,
        )

    def __Imop_ModalCoef_SetData(self, coef, index, size, pupil):
        return call_errmsg(
            imop_lib.Imop_ModalCoef_SetData,
            self.pointer,
            arr_p(coef),
            arr_p(index),
            size,
            pupil.pointer,
        )

    def __Imop_ModalCoef_GetData(self, n_modes, pupil):
        coef = np.zeros(n_modes, dtype=np.float32)
        index = np.zeros(n_modes, dtype=np.uint32)

        msg, result = call_errmsg(
            imop_lib.Imop_ModalCoef_GetData,
            self.pointer,
            arr_p(coef),
            arr_p(index),
            pupil.pointer,
        )

        return msg, result, coef, index

    def __Imop_ModalCoef_Delete(self):
        return call_errmsg(imop_lib.Imop_ModalCoef_Delete, self.pointer)


# CorrDataManager class (cCorrDataManager.h)
class CorrDataManager(Pointer):
    def __init__(self):
        super().__init__()

    def __del__(self):
        # Destructor
        self.__Imop_CorrDataManager_Delete()

    # Public functions:

    def new_from_backup_file(self, haso_config_file_path, interaction_matrix_file_path):
        self.__Imop_CorrDataManager_NewFromBackupFile(
            haso_config_file_path, interaction_matrix_file_path
        )

    def set_command_matrix_prefs(self, nb_kept_modes, tilt_filtering=False):
        self.__Imop_CorrDataManager_SetCommandMatrixPrefs(nb_kept_modes, tilt_filtering)

    def compute_command_matrix(self):
        self.__Imop_CorrDataManager_ComputeCommandMatrix()

    def get_greatest_common_pupil(self, pupil):
        _, _, pupil = self.__Imop_CorrDataManager_GetGreatestCommonPupil(pupil)

        return pupil

    def compute_delta_command_from_delta_slopes(self, delta_slopes, delta_command):
        (
            _,
            _,
            delta_command,
        ) = self.__Imop_CorrDataManager_ComputeDeltaCommandFromDeltaSlopes(
            delta_slopes, delta_command
        )

        return delta_command

    # Private functions translated directly from Imop C API:

    def __Imop_CorrDataManager_NewFromBackupFile(
        self, haso_config_file_path, interaction_matrix_file_path
    ):
        return call_errmsg(
            imop_lib.Imop_CorrDataManager_NewFromBackupFile,
            ct.byref(self.pointer),
            char_p(haso_config_file_path),
            char_p(interaction_matrix_file_path),
        )

    def __Imop_CorrDataManager_Delete(self):
        return call_errmsg(imop_lib.Imop_CorrDataManager_Delete, self.pointer)

    def __Imop_CorrDataManager_SetCommandMatrixPrefs(
        self, nb_kept_modes, tilt_filtering
    ):
        return call_errmsg(
            imop_lib.Imop_CorrDataManager_SetCommandMatrixPrefs,
            self.pointer,
            nb_kept_modes,
            tilt_filtering,
        )

    def __Imop_CorrDataManager_ComputeCommandMatrix(self):
        return call_errmsg(
            imop_lib.Imop_CorrDataManager_ComputeCommandMatrix, self.pointer
        )

    def __Imop_CorrDataManager_GetGreatestCommonPupil(self, pupil):
        pupil = pupil.astype(bool)

        msg, result = call_errmsg(
            imop_lib.Imop_CorrDataManager_GetGreatestCommonPupil,
            self.pointer,
            arr_p(pupil),
        )

        return msg, result, pupil

    def __Imop_CorrDataManager_ComputeDeltaCommandFromDeltaSlopes(
        self, delta_slopes, delta_command
    ):
        delta_command = delta_command.astype(np.float32)

        msg, result = call_errmsg(
            imop_lib.Imop_CorrDataManager_ComputeDeltaCommandFromDeltaSlopes,
            self.pointer,
            delta_slopes.pointer,  # HasoSlopes object
            arr_p(delta_command),
        )

        return msg, result, delta_command


# WavefrontCorrectorSet class (cWavefrontCorrectorSet.h)
class WavefrontCorrectorSet(Pointer):
    def __init__(self):
        super().__init__()

    def __del__(self):
        # Destructor
        self.__Imop_WavefrontCorrectorSet_Delete()

    # Public functions:

    def new_from_config_file(self, wfc_config_file_path):
        self.__Imop_WavefrontCorrectorSet_NewFromConfigFile(wfc_config_file_path)

    def get_actuators_count(self):
        _, _, n_actuators = self.__Imop_WavefrontCorrectorSet_GetActuatorsCount()

        return n_actuators

    # Private functions translated directly from Imop C API:

    def __Imop_WavefrontCorrectorSet_NewFromConfigFile(self, wfc_config_file_path):
        return call_errmsg(
            imop_lib.Imop_WavefrontCorrectorSet_NewFromConfigFile,
            ct.byref(self.pointer),
            char_p(wfc_config_file_path),
        )

    def __Imop_WavefrontCorrectorSet_GetActuatorsCount(self):
        n_actuators = ct.c_int()

        msg, result = call_errmsg(
            imop_lib.Imop_WavefrontCorrectorSet_GetActuatorsCount,
            self.pointer,
            ct.byref(n_actuators),
        )

        return msg, result, n_actuators.value

    def __Imop_WavefrontCorrectorSet_Delete(self):
        return call_errmsg(imop_lib.Imop_WavefrontCorrectorSet_Delete, self.pointer)


# WavefrontCorrector class (cWavefrontCorrector.h)
class WavefrontCorrector(Pointer):
    def __init__(
        self,
        wfc_config_file_path=os.path.join(
            basepath, "MirrorFiles", "WaveFrontCorrector_Mirao52-e_0259.dat"
        ),
        haso_config_file_path=os.path.join(
            basepath, "MirrorFiles", "HASO4_first_7458.dat"
        ),
        positions_file_path=os.path.join(
            basepath, "MirrorFiles", "FlouresceinOctober12.wcs"
        ),
        interaction_matrix_file_path=os.path.join(
            basepath, "MirrorFiles", "OlympusApril22.aoc"
        ),
        n_actuators=52,
    ):
        super().__init__()

        self.wfc_config_file_path = wfc_config_file_path
        self.haso_config_file_path = haso_config_file_path
        self.positions_file_path = positions_file_path
        self.interaction_matrix_file_path = interaction_matrix_file_path

        self.n_actuators = n_actuators

        self.__Imop_WavefrontCorrector_NewFromConfigFile()
        self.__Imop_WavefrontCorrector_Init()

        self.get_preferences()

    def __del__(self):
        # Destructor
        self.__Imop_WavefrontCorrector_Delete()

    # Public functions:

    def get_preferences(self):
        (
            _,
            _,
            (sleep_after_movement, cmd_min, cmd_max, validity, fixed_values),
        ) = self.__Imop_WavefrontCorrector_GetPreferences()

        self.preferences = {
            "sleep_after_movement": sleep_after_movement,
            "cmd_min": cmd_min,
            "cmd_max": cmd_max,
            "validity": validity,
            "fixed_values": fixed_values,
        }

        return self.preferences

    def set_preferences(self, **kwargs):
        for k in kwargs.keys():
            if k in self.preferences.keys():
                self.preferences[k] = kwargs[k]
            else:
                raise ValueError(
                    k
                    + " is not a valid preference! Choose from: "
                    + str([k for k in self.preferences.keys()])
                )
                return

        self.__Imop_WavefrontCorrector_SetPreferences(
            self.preferences["sleep_after_movement"],
            self.preferences["cmd_min"],
            self.preferences["cmd_max"],
            self.preferences["validity"],
            self.preferences["fixed_values"],
        )

    def get_temporization(self):
        _, _, sleep_after_movement = self.__Imop_WavefrontCorrector_GetTemporization()

        return sleep_after_movement

    def set_temporization(self, sleep_after_movement):
        self.preferences["sleep_after_movement"] = sleep_after_movement

        self.__Imop_WavefrontCorrector_SetTemporization(sleep_after_movement)

    def get_current_positions(self):
        _, _, positions = self.__Imop_WavefrontCorrector_GetCurrentPositions()

        return positions

    def get_positions_from_file(self, path):
        _, _, positions = self.__Imop_WavefrontCorrector_GetPositionsFromFile(path)

        return positions

    def save_positions_to_file(self, file_path):
        self.__Imop_WavefrontCorrector_SaveCurrentPositionsToFile(file_path)

    def move_relative(self, positions):
        self.__Imop_WavefrontCorrector_MoveToRelativePositions(
            positions.astype(np.float32)
        )

    def move_absolute(self, positions):
        self.__Imop_WavefrontCorrector_MoveToAbsolutePositions(
            positions.astype(np.float32)
        )

    # TODO: doesn't really work... always returns zero.
    def check_absolute_positions(self, positions):
        _, flag = self.__Imop_WavefrontCorrector_CheckAbsolutePositions(
            positions.astype(np.float32)
        )
        return flag

    def clear(self):
        self.move_absolute(np.zeros(self.n_actuators))

    # Private functions translated directly from Imop C API:

    def __Imop_WavefrontCorrector_Delete(self):
        return call_errmsg(imop_lib.Imop_WavefrontCorrector_Delete, self.pointer)

    def __Imop_WavefrontCorrector_NewFromConfigFile(self):
        return call_errmsg(
            imop_lib.Imop_WavefrontCorrector_NewFromConfigFile,
            ct.byref(self.pointer),
            char_p(self.wfc_config_file_path),
        )

    def __Imop_WavefrontCorrector_Init(self, set_init_state_from_config_file=True):
        return call_errmsg(
            imop_lib.Imop_WavefrontCorrector_Init,
            self.pointer,
            ct.c_bool(set_init_state_from_config_file),
        )

    def __Imop_WavefrontCorrector_CallSpecificFeature(self, feature_name):
        return call_errmsg(
            imop_lib.Imop_WavefrontCorrector_CallSpecificFeature,
            self.pointer,
            char_p(feature_name),
        )

    def __Imop_WavefrontCorrector_GetPreferences(self):
        sleep_after_movement = ct.c_int()
        cmd_min = np.zeros(self.n_actuators).astype(np.float32)
        cmd_max = np.zeros(self.n_actuators).astype(np.float32)
        validity = np.zeros(self.n_actuators).astype(np.int32)
        fixed_values = np.zeros(self.n_actuators).astype(np.float32)

        msg, result = call_errmsg(
            imop_lib.Imop_WavefrontCorrector_GetPreferences,
            self.pointer,
            ct.byref(sleep_after_movement),
            arr_p(cmd_min),
            arr_p(cmd_max),
            arr_p(validity),
            arr_p(fixed_values),
        )

        return (
            msg,
            result,
            (sleep_after_movement.value, cmd_min, cmd_max, validity, fixed_values),
        )

    # TODO: Imop_WavefrontCorrector_AssertEqualPreferences

    # TODO: Imop_WavefrontCorrector_CheckUserPreferences

    def __Imop_WavefrontCorrector_SetPreferences(
        self, sleep_after_movement, cmd_min, cmd_max, validity, fixed_values
    ):
        return call_errmsg(
            imop_lib.Imop_WavefrontCorrector_SetPreferences,
            self.pointer,
            sleep_after_movement,
            arr_p(cmd_min),
            arr_p(cmd_max),
            arr_p(validity),
            arr_p(fixed_values),
        )

    def __Imop_WavefrontCorrector_GetTemporization(self):
        sleep_after_movement = ct.c_int()

        msg, result = call_errmsg(
            imop_lib.Imop_WavefrontCorrector_GetTemporization,
            self.pointer,
            ct.byref(sleep_after_movement),
        )

        return msg, result, sleep_after_movement.value

    def __Imop_WavefrontCorrector_SetTemporization(self, sleep_after_movement):
        return call_errmsg(
            imop_lib.Imop_WavefrontCorrector_SetTemporization,
            self.pointer,
            sleep_after_movement,
        )

    def __Imop_WavefrontCorrector_GetCurrentPositions(self):
        positions = np.zeros(self.n_actuators).astype(np.float32)

        msg, result = call_errmsg(
            imop_lib.Imop_WavefrontCorrector_GetCurrentPositions,
            self.pointer,
            arr_p(positions),
        )

        return msg, result, positions

    # TODO: Imop_WavefrontCorrector_CheckRelativePositions

    def __Imop_WavefrontCorrector_MoveToRelativePositions(self, positions):
        return call_errmsg(
            imop_lib.Imop_WavefrontCorrector_MoveToRelativePositions,
            self.pointer,
            arr_p(positions),
        )

    # TODO: Imop_WavefrontCorrector_CheckAbsolutePositions
    def __Imop_WavefrontCorrector_CheckAbsolutePositions(self, positions):
        msg, result = call_errmsg(
            imop_lib.Imop_WavefrontCorrector_CheckAbsolutePositions,
            self.pointer,
            arr_p(positions),
        )

        return msg, result

    def __Imop_WavefrontCorrector_MoveToAbsolutePositions(self, positions):
        return call_errmsg(
            imop_lib.Imop_WavefrontCorrector_MoveToAbsolutePositions,
            self.pointer,
            arr_p(positions),
        )

    def __Imop_WavefrontCorrector_GetPositionsFromFile(self, pmc_file_path):
        positions = np.zeros(self.n_actuators).astype(np.float32)

        msg, result = call_errmsg(
            imop_lib.Imop_WavefrontCorrector_GetPositionsFromFile,
            char_p(pmc_file_path),
            arr_p(positions),
        )

        return msg, result, positions

    def __Imop_WavefrontCorrector_SaveCurrentPositionsToFile(self, pmc_file_path):
        return call_errmsg(
            imop_lib.Imop_WavefrontCorrector_SaveCurrentPositionsToFile,
            self.pointer,
            char_p(pmc_file_path),
        )

    # TODO: Rest of file... but maybe good enough for now.



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import importlib
from threading import Lock

def combine_funcs(*funclist):
    """this function will combine a list of functions to a new function

    Parameters
    ----------
    funclist: list
        a list of functions

    Returns
    -------
    new_func: function
        a new function
    """

    def new_func():
        for func in funclist:
            if callable(func):
                func()

    return new_func


def build_ref_name(separator, *args):
    """this function will build a reference name

    Parameters
    ----------
    separator: str
        the separator between each argument
    args: list
        a list of arguments

    Returns
    -------
    ref_name: str
        the reference name
    """
    alist = list(map(lambda a: str(a), args))
    return separator.join(alist)


def copy_proxy_object(content):
    """This function will serialize proxy dict and list

    Parameters
    ----------
    content: dict/list
        the proxy object

    Returns
    -------
    result: dict/list
    """
    from multiprocessing import managers

    def func(content):
        if type(content) == managers.DictProxy:
            result = {}
            for k in content.keys():
                result[k] = func(content[k])
        elif type(content) == managers.ListProxy:
            result = []
            for v in content:
                result.append(func(v))
        else:
            result = content
        return result

    return func(content)


def load_module_from_file(module_name, file_path):
    """This function will load python file from file path as a module

    Parameters
    ----------
    module_name: str
        the module name
    file_path: os.path/str
        the python file path

    Returns
    -------
    module: class of module
    """
    try:
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except ModuleNotFoundError:
        return None
    return module

class VariableWithLock:
    def __init__(self, VariableType):
        self.lock = Lock()
        self.value = VariableType()

    def __enter__(self):
        self.lock.acquire()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging
from tkinter import messagebox

# Third Party Imports

# Local Imports
from navigate.tools.multipos_table_tools import (
    sign,
    compute_tiles_from_bounding_box,
    calc_num_tiles,
    update_table,
)
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.tools.common_functions import combine_funcs

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class TilingWizardController(GUIController):
    """Tiling Wizard Controller

    Controller for tiling wizard parameters.
    Gathers the FOV from the camera settings tab and will
    update when user changes this value.
    Set start/end position buttons will grab the stage
    values for the respective axis when pressed and display in popup
    Number of images we need to acquire with our desired
    percent overlap is calculated and then displayed in third column
    """

    def __init__(self, view, parent_controller):
        """Initialize Tiling Wizard Controller

        Parameters
        ----------
        view : object
            GUI element containing widgets and variables to control.
            Likely tk.Toplevel-derived. In this case tiling_wizard_popup.py
        parent_controller : channels_tab_controller
            The controller that creates the popup/this controller.

        """
        super().__init__(view, parent_controller)

        # Getting widgets and buttons and vars of widgets
        #: dict: Dictionary of widgets in the view
        self.widgets = self.view.get_widgets()
        #: dict: Dictionary of buttons in the view
        self.buttons = self.view.get_buttons()
        #: dict: Dictionary of variables in the view
        self.variables = self.view.get_variables()
        #: int: Default percent overlap between tiles
        self._percent_overlap = 10.0  # default to 10% overlap
        #: dict: flags indicating if all the value are correct to set the table
        self.is_validated = {"x": True, "y": True, "z": True, "f": True}

        # Init widgets to zero
        #: list: List of axes to iterate over
        self._axes = ["x", "y", "z", "f"]
        self.variables["percent_overlap"].set(self._percent_overlap)
        self.variables["total_tiles"].set(1)
        for ax in self._axes:
            self.variables[f"{ax}_start"].set(0.0)
            self.variables[f"{ax}_end"].set(0.0)
            self.variables[f"{ax}_dist"].set(0.0)
            self.variables[f"{ax}_fov"].set(0.0)
            self.variables[f"{ax}_tiles"].set(1)

        # Ref to widgets in other views
        # (Camera Settings, Stage Control Positions, Stack Acq Settings)
        main_view = (
            self.parent_controller.parent_controller.view
        )  # channels_tab_controller -> navigate_controller -> view
        self.cam_settings_widgets = (
            main_view.settings.camera_settings_tab.camera_roi.get_widgets()
        )
        self.stack_acq_widgets = (
            main_view.settings.channels_tab.stack_acq_frame.get_widgets()
        )
        self.stage_position_vars = (
            main_view.settings.stage_control_tab.position_frame.get_variables()
        )
        self.multipoint_table = (
            main_view.settings.multiposition_tab.multipoint_list.get_table()
        )

        # Setting/Tracing Percent Overlay
        # Overlay change is also handled in update_overlap
        self.variables["percent_overlap"].trace_add(
            "write", lambda *args: self.update_overlap()
        )

        # Trace cam_settings FOV to catch user changes
        # FOV change handled in update_fov
        self.cam_settings_widgets["FOV_X"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("y")
        )
        self.cam_settings_widgets["FOV_Y"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("x")
        )
        self.stack_acq_widgets["abs_z_start"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("z")
        )
        self.stack_acq_widgets["abs_z_end"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("z")
        )
        self.stack_acq_widgets["start_focus"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("f")
        )
        self.stack_acq_widgets["end_focus"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("f")
        )

        # Calculate distances
        self.variables["x_start"].trace_add(
            "write", lambda *args: self.calculate_distance("x")
        )
        self.variables["x_end"].trace_add(
            "write", lambda *args: self.calculate_distance("x")
        )
        self.variables["y_start"].trace_add(
            "write", lambda *args: self.calculate_distance("y")
        )
        self.variables["y_end"].trace_add(
            "write", lambda *args: self.calculate_distance("y")
        )
        self.variables["z_start"].trace_add(
            "write", lambda *args: self.calculate_distance("z")
        )
        self.variables["z_end"].trace_add(
            "write", lambda *args: self.calculate_distance("z")
        )
        self.variables["f_start"].trace_add(
            "write", lambda *args: self.calculate_distance("f")
        )
        self.variables["f_end"].trace_add(
            "write", lambda *args: self.calculate_distance("f")
        )

        # Bind FOV changes
        self.variables["x_fov"].trace_add(
            "write", lambda *args: self.calculate_tiles("x")
        )
        self.variables["y_fov"].trace_add(
            "write", lambda *args: self.calculate_tiles("y")
        )
        self.variables["z_fov"].trace_add(
            "write", lambda *args: self.calculate_tiles("z")
        )
        self.variables["f_fov"].trace_add(
            "write", lambda *args: self.calculate_tiles("f")
        )

        self.variables["x_start"].trace_add("write", lambda *args: self.update_fov("x"))
        self.variables["x_end"].trace_add("write", lambda *args: self.update_fov("x"))
        self.variables["y_start"].trace_add("write", lambda *args: self.update_fov("y"))
        self.variables["y_end"].trace_add("write", lambda *args: self.update_fov("y"))
        self.variables["z_start"].trace_add("write", lambda *args: self.update_fov("z"))
        self.variables["z_end"].trace_add("write", lambda *args: self.update_fov("z"))
        self.variables["f_start"].trace_add("write", lambda *args: self.update_fov("f"))
        self.variables["f_end"].trace_add("write", lambda *args: self.update_fov("f"))

        # Calculating Number of Tiles traces
        self.variables["x_dist"].trace_add(
            "write", lambda *args: self.calculate_tiles("x")
        )
        self.variables["y_dist"].trace_add(
            "write", lambda *args: self.calculate_tiles("y")
        )
        self.variables["z_dist"].trace_add(
            "write", lambda *args: self.calculate_tiles("z")
        )
        self.variables["f_dist"].trace_add(
            "write", lambda *args: self.calculate_tiles("f")
        )

        # Populate Table trace
        self.buttons["set_table"].configure(command=self.set_table)

        for ax in self._axes:
            # Start/End buttons
            self.buttons[f"{ax}_start"].configure(
                command=self.position_handler(ax, "start")
            )
            self.buttons[f"{ax}_end"].configure(
                command=self.position_handler(ax, "end")
            )

            # Calculating total tile traces
            self.variables[f"{ax}_tiles"].trace_add(
                "write", lambda *args: self.update_total_tiles()
            )

            # Update widgets to current values in other views
            self.update_fov(ax)

        # Properly Closing Popup with parent controller
        self.view.popup.protocol(
            "WM_DELETE_WINDOW",
            combine_funcs(
                self.view.popup.dismiss,
                lambda: delattr(self.parent_controller, "tiling_wizard_controller"),
            ),
        )

    def set_table(self):
        """Set the multipoint table to the values in the tiling wizard

        Sets multiposition table with values from tiling wizard after
        populate Multiposition Table button is pressed
        Compute grid will return a list of all position combinations.
        This list is then converted to a
        pandas dataframe which is then set as the new table data.
        The table is then redrawn.

        Examples
        --------
        >>> self.set_table()
        """
        if False in self.is_validated.values():
            messagebox.showwarning(
                title="Navigate",
                message="Can't calculate positions, please make sure all FOV Dists are correct!",
            )
            return

        x_start = float(self.variables["x_start"].get())
        x_stop = float(self.variables["x_end"].get())
        x_tiles = int(self.variables["x_tiles"].get())

        y_start = float(self.variables["y_start"].get())
        y_stop = float(self.variables["y_end"].get())
        y_tiles = int(self.variables["y_tiles"].get())

        # shift z by coordinate origin of local z-stack
        z_start = float(self.variables["z_start"].get()) - float(
            self.stack_acq_widgets["start_position"].get()
        )
        z_stop = float(self.variables["z_end"].get()) - float(
            self.stack_acq_widgets["end_position"].get()
        )
        z_tiles = int(self.variables["z_tiles"].get())

        # Default to fixed theta
        r_start = self.stage_position_vars["theta"].get()
        r_stop = self.stage_position_vars["theta"].get()
        r_tiles = 1

        f_start = float(self.variables["f_start"].get()) - float(
            self.stack_acq_widgets["start_focus"].get()
        )
        f_stop = float(self.variables["f_end"].get()) - float(
            self.stack_acq_widgets["end_focus"].get()
        )
        f_tiles = int(self.variables["f_tiles"].get())

        # for consistency, always go from low to high
        def sort_vars(a, b):
            """Sort two variables from low to high

            Parameters
            ----------
            a : float
                First variable
            b : float
                Second variable

            Returns
            -------
            a, b : float
                Sorted variables
            """
            if a > b:
                return b, a
            return a, b

        x_start, x_stop = sort_vars(x_start, x_stop)
        y_start, y_stop = sort_vars(y_start, y_stop)
        z_start, z_stop = sort_vars(z_start, z_stop)
        r_start, r_stop = sort_vars(r_start, r_stop)
        f_start, f_stop = sort_vars(f_start, f_stop)

        ov = float(self._percent_overlap) / 100
        table_values = compute_tiles_from_bounding_box(
            x_start,
            x_tiles,
            float(self.variables["x_fov"].get()),
            ov,
            y_start,
            y_tiles,
            float(self.variables["y_fov"].get()),
            ov,
            z_start,
            z_tiles,
            float(self.variables["z_fov"].get()),
            ov,
            r_start,
            r_tiles,
            0,
            ov,
            f_start,
            f_tiles,
            float(self.variables["f_fov"].get()),
            ov,
        )

        update_table(self.multipoint_table, table_values)

        # If we have additional axes, create self.d{axis} for each
        # additional axis, to ensure we keep track of the step size
        config = self.parent_controller.parent_controller.configuration
        microscope_name = config["experiment"]["MicroscopeState"]["microscope_name"]
        scope = config["configuration"]["microscopes"][microscope_name]
        coupled_axes = scope["stage"].get("coupled_axes", None)
        if coupled_axes is not None:
            for follower in coupled_axes.values():
                config["experiment"]["MicroscopeState"][
                    f"{follower.lower()}_step_size"
                ] = self.variables[f"{follower.lower()}_fov"].get()

    def update_total_tiles(self):
        """Update the total number of tiles in the tiling wizard

        Sums the tiles for each axis in the tiling wizard.
        Will update when any axis has a tile amount change.

        Examples
        --------
        >>> self.update_total_tiles()
        """

        total_tiles = 1
        for ax in self._axes:
            total_tiles *= float(self.variables[f"{ax}_tiles"].get())

        self.variables["total_tiles"].set(total_tiles)

    def calculate_tiles(self, axis=None):
        """Calculate the number of tiles for a given axis

        Calculates the number of tiles of the acquisition for
        each axis or an individual axis
        Num of Tiles = dist - (overlay * FOV)  /  FOV * (1 - overlay)
        (D-OF)/(F-OF) = N

        Parameters
        ----------
        axis : str
            x, y, z, f axis of stage to calculate.

        Examples
        --------
        >>> self.calculate_tiles()
        """

        if axis not in self._axes + [None]:
            logger.warning(
                f"Controller - Tiling Wizard - Unknown axis {axis}, "
                "skipping calculate_tiles()."
            )
            return

        if axis is not None:
            if not isinstance(axis, list):
                axis = [axis]
        else:
            axis = self._axes

        overlay = float(self._percent_overlap) / 100

        for ax in axis:
            self.is_validated[ax] = True
            fov_value = self.variables[f"{ax}_fov"].get()
            if not fov_value or "inf" in fov_value:
                self.is_validated[ax] = False
                return
            try:
                dist = abs(float(self.variables[f"{ax}_dist"].get()))  # um
                fov = abs(float(self.variables[f"{ax}_fov"].get()))  # um

                if ax.lower() == "x" or ax.lower() == "y":
                    # + fov because distance is center of the fov to center of the fov
                    # and so we are covering a distance that is 2 * 1/2 * fov
                    # larger than dist
                    dist += fov

                num_tiles = calc_num_tiles(dist, overlay, fov)

                self.variables[f"{ax}_tiles"].set(num_tiles)
            except ValueError as e:
                self.is_validated[ax] = False
                logger.warning(f"Controller - Tiling Wizard - {e}")

    def calculate_distance(self, axis):
        """Calculate the distance for a given axis

        This function will calculate the distance for a given
        axis of the stage when the start or end position is changed
        via the Set buttons

        Parameters
        ----------
        axis : str
            x, y, z axis of stage to calculate

        Examples
        --------
        >>> self.calculate_distance()
        """

        start = float(self.variables[axis + "_start"].get())
        end = float(self.variables[axis + "_end"].get())
        dist = abs(end - start)
        self.variables[axis + "_dist"].set(dist)

    def update_overlap(self):
        """Update the overlay percentage for the tiling wizard

        Updates percent overlay when a user changes the widget in the popup.
        This value is used for backend calculations.
        The number of tiles will then be recalculated

        Examples
        --------
        >>> self.update_overlap()
        """

        try:
            self._percent_overlap = float(self.variables["percent_overlap"].get())
            self.calculate_tiles()
        except ValueError:
            # most likely an empty string was passed
            pass

    def position_handler(self, axis, start_end):
        """Set the start or end position for a given axis

        When the Set [axis] Start/End button is pressed then the
        stage position is polled from the stage controller

        Parameters
        ----------
        axis : str
            x, y, z, f axis that corresponds to stage axis
        start_end : str
            start or end will signify which spinbox gets updated upon button press

        Returns
        -------
        handler : func
            Function for setting positional spinbox based on parameters passed in

        Examples
        --------
        >>> self.position_handler()
        """

        def handler():
            # Force us to get the current stage positions from the stage
            self.parent_controller.parent_controller.execute("stop_stage")

            def set_bounds(axis, start_end):
                # Now set the bounds
                pos = self.stage_position_vars[axis].get()
                self.widgets[axis + "_" + start_end].widget.set(pos)
                # if axis == "z":
                #     setattr(self, f"_f_{start_end}",
                #             self.stage_position_vars["f"].get())

            self.parent_controller.parent_controller.view.after(
                250, lambda: set_bounds(axis, start_end)
            )

        return handler

    def update_fov(self, axis=None):
        """Update the FOV for the tiling wizard

        Grabs the updated FOV if changed by user,
        will recalculate num of tiles for each axis after

        Parameters
        ----------
        axis : str
            Axis

        Examples
        --------
        >>> self.update_fov()
        """

        if axis is None:
            axis = self._axes
        elif isinstance(axis, str):
            axis = [axis]

        for ax in axis:
            try:
                # Calculate signed fov
                if ax == "y":
                    y = float(self.cam_settings_widgets["FOV_X"].get()) * sign(
                        float(self.variables["x_end"].get())
                        - float(self.variables["x_start"].get())
                    )
                elif ax == "x":
                    x = float(self.cam_settings_widgets["FOV_Y"].get()) * sign(
                        float(self.variables["y_end"].get())
                        - float(self.variables["y_start"].get())
                    )
                elif ax == "z":
                    z = float(self.stack_acq_widgets["end_position"].get()) - float(
                        self.stack_acq_widgets["start_position"].get()
                    )
                elif ax == "f":
                    f = float(self.stack_acq_widgets["end_focus"].get()) - float(
                        self.stack_acq_widgets["start_focus"].get()
                    )

                # for ax in self._axes:
                # self._fov[ax] = locals().get(ax)
                self.variables[f"{ax}_fov"].set(
                    abs(locals().get(ax))
                )  # abs(self._fov[ax]))

                self.calculate_tiles(ax)
            except ValueError as e:
                logger.debug(
                    f"Controller - Tiling Wizard - Caught ValueError: {e}. "
                    "Declining to update FOV."
                )
                pass

    def showup(self):
        """Show the tiling wizard

        Brings popup window to front of screen

        Examples
        --------
        >>> self.showup()
        """
        self.view.popup.deiconify()
        self.view.popup.attributes("-topmost", 1)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import importlib
import logging
import time

from navigate.model.devices.stages.stage_base import StageBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


def build_MCLStage_connection(serialnum):
    """Build a connection to the Mad City Lab stage.

    Parameters
    ----------
    serialnum : int
        Serial number of the stage.

    Returns
    -------
    stage_connection : dict
        Dictionary containing the connection information for the stage.
    """
    mcl_controller = importlib.import_module("navigate.model.devices.APIs.mcl.madlib")

    # Initialize
    mcl_controller.MCL_GrabAllHandles()

    handle = mcl_controller.MCL_GetHandleBySerial(int(serialnum))

    stage_connection = {"handle": handle, "controller": mcl_controller}

    return stage_connection


class MCLStage(StageBase):
    """Mad City Lab stage class."""

    def __init__(self, microscope_name, device_connection, configuration, device_id=0):
        """Initialize the MCL stage.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : dict
            Dictionary containing the connection information for the device.
        configuration : dict
            Dictionary containing the configuration information for the device.
        device_id : int
            Device ID.
        """
        super().__init__(
            microscope_name, device_connection, configuration, device_id
        )  # only initialize the focus axis

        # Mapping from self.axes to corresponding MCL channels
        if device_connection is not None:
            #: object: MCL controller object.
            self.mcl_controller = device_connection["controller"]

            #: int: MCL handle.
            self.handle = device_connection["handle"]

        # Default axes mapping
        # the API maps axis to a number
        axes_mapping = {"x": "x", "y": "y", "z": "z", "f": "f", "theta": "aux"}
        if not self.axes_mapping:
            #: dict: Dictionary of software axes and their corresponding hardware axes.
            self.axes_mapping = {
                axis: axes_mapping[axis] for axis in self.axes if axis in axes_mapping
            }

    def report_position(self):
        """Report the position of the stage.

        Reports the position of the stage for all axes, and creates the hardware
        position dictionary.
        """
        for ax in self.axes_mapping:
            try:
                pos = self.mcl_controller.MCL_SingleReadN(
                    self.axes_mapping[ax], self.handle
                )
                setattr(self, f"{ax}_pos", pos)
            except self.mcl_controller.MadlibError as e:
                logger.debug(f"MCL - {e}")
                pass

        return self.get_position_dict()

    def move_axis_absolute(self, axis, abs_pos, wait_until_done=False):
        """Implement movement logic along a single axis.

        Example calls:

        Parameters
        ----------
        axis : str
            An axis. For example, 'x', 'y', 'z', 'f', 'theta'.
        abs_pos : float
            Absolute position value
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        bool
            Was the move successful?
        """

        axis_abs = self.get_abs_position(axis, abs_pos)
        if axis_abs == -1e50:
            return False

        self.mcl_controller.MCL_SingleWriteN(
            axis_abs, self.axes_mapping[axis], self.handle
        )

        if wait_until_done:
            stage_pos, n_tries, i = -1e50, 10, 0
            while (abs(stage_pos - abs_pos) < 0.01) and (i < n_tries):
                stage_pos = self.mcl_controller.MCL_SingleReadN(axis, self.handle)
                i += 1
                time.sleep(0.01)
            if abs(stage_pos - abs_pos) > 0.01:
                return False
        return True

    def move_absolute(self, move_dictionary, wait_until_done=False):
        """Move the stage to an absolute position.

        Parameters
        ----------
        move_dictionary : dict
            A dictionary of values required for movement. Includes 'x_abs', etc. for
            one or more axes. Expects values in micrometers, except for theta, which is
            in degrees.
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        success : bool
            Was the move successful?
        """
        abs_pos_dict = self.verify_abs_position(move_dictionary)
        if not abs_pos_dict:
            return False

        result = True
        for ax in abs_pos_dict:
            success = self.move_axis_absolute(ax, abs_pos_dict[ax], wait_until_done)
            result = result and success

        return result



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
from tkinter import ttk, Grid
import logging

# Third Party Imports
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

# Local Imports
from navigate.view.custom_widgets.DockableNotebook import DockableNotebook
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput


# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class CameraNotebook(DockableNotebook):
    """This class is the notebook that holds the camera view and waveform settings
    tabs."""

    def __init__(self, frame_top_right, *args, **kwargs):
        """Init function for the CameraNotebook class.


        Parameters
        ----------
        frame_top_right : tk.Frame
            The frame that will hold the notebook.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        # Init notebook
        DockableNotebook.__init__(self, frame_top_right, *args, **kwargs)

        # Putting notebook 2 into top right frame
        self.grid(row=0, column=0)

        #: CameraTab: The camera tab.
        self.camera_tab = CameraTab(self)

        #: WaveformTab: The waveform settings tab.
        self.waveform_tab = WaveformTab(self)

        # Tab list
        tab_list = [self.camera_tab, self.waveform_tab]
        self.set_tablist(tab_list)

        # Adding tabs to self notebook
        self.add(self.camera_tab, text="Camera View", sticky=tk.NSEW)
        self.add(self.waveform_tab, text="Waveform Settings", sticky=tk.NSEW)


class CameraTab(tk.Frame):
    """CameraTab class."""

    def __init__(self, cam_wave, *args, **kwargs):
        """Initialize the CameraTab class.

        Parameters
        ----------
        cam_wave : tk.Frame
            The frame that will hold the camera tab.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        #  Init Frame
        tk.Frame.__init__(self, cam_wave, *args, **kwargs)

        #: int: The index of the tab.
        self.index = 0

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: tk.Frame: The frame that will hold the camera image.
        self.cam_image = ttk.Frame(self)
        self.cam_image.grid(row=0, column=0, rowspan=3, sticky=tk.NSEW)

        #: Bool: The popup flag.
        self.is_popup = False

        #: Bool: The docked flag.
        self.is_docked = True

        #: int: The width of the canvas.
        #: int: The height of the canvas.
        self.canvas_width, self.canvas_height = 512, 512

        #: tk.Canvas: The canvas that will hold the camera image.
        self.canvas = tk.Canvas(
            self.cam_image, width=self.canvas_width, height=self.canvas_height
        )
        self.canvas.grid(row=0, column=0, sticky=tk.NSEW, padx=5, pady=5)

        #: matplotlib.figure.Figure: The figure that will hold the camera image.
        self.matplotlib_figure = Figure(figsize=[6, 6], tight_layout=True)

        #: matplotlib.backends.backend_tkagg.FigureCanvasTkAgg: The canvas that will
        # hold the camera image.
        self.matplotlib_canvas = FigureCanvasTkAgg(self.matplotlib_figure, self.canvas)

        #: IntensityFrame: The frame that will hold the scale settings/palette color.
        self.scale_palette = IntensityFrame(self)
        self.scale_palette.grid(row=0, column=1, sticky=tk.NSEW, padx=5, pady=5)

        #: tk.Scale: The slider that will hold the slice index.
        self.slider = tk.Scale(
            self,
            from_=0,
            to=200,
            tickinterval=20,
            orient=tk.HORIZONTAL,
            showvalue=0,
            label="Slice Index",
        )
        self.slider.configure(state="disabled")  # 'normal'
        self.slider.grid(row=3, column=0, sticky=tk.NSEW, padx=5, pady=5)

        #: MetricsFrame: The frame that will hold the camera selection and counts.
        self.image_metrics = MetricsFrame(self)
        self.image_metrics.grid(row=1, column=1, sticky=tk.NSEW, padx=5, pady=5)

        #: RenderFrame: The frame that will hold the live display functionality.
        self.live_frame = RenderFrame(self)
        self.live_frame.grid(row=2, column=1, sticky=tk.NSEW, padx=5, pady=5)


class RenderFrame(ttk.Labelframe):
    """This class is the frame that holds the live display functionality."""

    def __init__(self, cam_view, *args, **kwargs):
        """Initialize the RenderFrame class.

        Parameters
        ----------
        cam_view : tk.Frame
            The frame that will hold the live display functionality.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        # Init Frame
        text_label = "Image Display"
        ttk.Labelframe.__init__(self, cam_view, text=text_label, *args, **kwargs)

        # Formatting
        Grid.columnconfigure(self, "all", weight=1)
        Grid.rowconfigure(self, "all", weight=1)

        #: tk.StringVar: The variable that holds the live display functionality.
        self.live_var = tk.StringVar()

        #: ttk.Combobox: The combobox that holds the live display functionality.
        self.live = ttk.Combobox(
            self, textvariable=self.live_var, state="readonly", width=6
        )
        self.live["values"] = (
            "Live",
            "XY Slice",
            "YZ Slice",
            "ZY Slice",
            "XY MIP",
            "YZ MIP",
            "ZY MIP",
        )
        self.live.set("Live")
        self.live.grid(row=0, column=0)
        self.live.state = "readonly"

    def get_variables(self):
        """Function to get the variables.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        variables : dict
            The dictionary that holds the variables.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get()
        return variables

    def get_widgets(self):
        """Function to get the widgets.

        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        self.inputs : dict
            The dictionary that holds the widgets.
        """
        return self.inputs


class WaveformTab(tk.Frame):
    """This class is the frame that holds the waveform tab."""

    def __init__(self, cam_wave, *args, **kwargs):
        """Initialize the WaveformTab class.

        Parameters
        ----------
        cam_wave : tk.Frame
            The frame that will hold the waveform tab.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.

        """
        # Init Frame
        tk.Frame.__init__(self, cam_wave, *args, **kwargs)

        #: int: The index of the tab.
        self.index = 1

        #: bool: The popup flag.
        self.is_docked = True

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: tk.Frame: The frame that will hold the waveform plots.
        self.waveform_plots = ttk.Frame(self)
        self.waveform_plots.grid(row=0, column=0, sticky=tk.NSEW)

        #: matplotlib.figure.Figure: The figure that will hold the waveform plots.
        self.fig = Figure(figsize=(6, 6), dpi=100)

        #: matplotlib.backends.backend_tkagg.FigureCanvasTkAgg: The canvas that will
        # hold the waveform plots.
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.waveform_plots)
        self.canvas.draw()

        #: tk.Frame: The frame that will hold the waveform settings.
        self.waveform_settings = WaveformSettingsFrame(self)
        self.waveform_settings.grid(row=1, column=0, sticky=tk.NSEW, padx=5, pady=5)


class WaveformSettingsFrame(ttk.Labelframe):
    """This class is the frame that holds the waveform settings."""

    def __init__(self, wav_view, *args, **kwargs):
        """Initialize the WaveformSettingsFrame class.

        Parameters
        ----------
        wav_view : tk.Frame
            The frame that will hold the waveform settings.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        # Init Frame
        text_label = "Settings"
        ttk.Labelframe.__init__(self, wav_view, text=text_label, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: dict: The dictionary that holds the widgets.
        self.inputs = {}

        self.inputs["sample_rate"] = LabelInput(
            parent=self,
            label="Sample rate",
            input_class=ttk.Spinbox,
            input_var=tk.IntVar(),
            input_args={"from_": 1, "to": 2**16 - 1, "increment": 1, "width": 5},
        )
        self.inputs["sample_rate"].grid(row=0, column=0, sticky=tk.NSEW, padx=3, pady=3)

        self.inputs["waveform_template"] = LabelInput(
            parent=self,
            label="Waveform Template",
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            input_args={"width": 20},
        )
        self.inputs["waveform_template"].grid(
            row=0, column=1, sticky=tk.NSEW, padx=3, pady=3
        )

    def get_variables(self):
        """Function to get the variables.

        Returns
        -------
        variables : dict
            The dictionary that holds the variables.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get()
        return variables

    def get_widgets(self):
        """Function to get the widgets.

        Returns
        -------
        self.inputs : dict
            The dictionary that holds the widgets.
        """
        return self.inputs


class MetricsFrame(ttk.Labelframe):
    """This class is the frame that holds the image metrics."""

    def __init__(self, cam_view, *args, **kwargs):
        """Initialize the MetricsFrame class.

        Parameters
        ----------
        cam_view : tk.Frame
            The frame that will hold the image metrics.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        # Init Labelframe
        text_label = "Image Metrics"
        ttk.Labelframe.__init__(self, cam_view, text=text_label, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        # Dictionary for widgets
        #: dict: The dictionary that holds the widgets.
        self.inputs = {}

        # Labels and names
        #: list: The list of labels for the widgets.
        self.labels = ["Frames to Avg", "Image Max Counts", "Channel"]
        #: list: The list of names for the widgets.
        self.names = ["Frames", "Image", "Channel"]

        # Loop for widgets
        for i in range(len(self.labels)):
            if i == 0:
                self.inputs[self.names[i]] = LabelInput(
                    parent=self,
                    label=self.labels[i],
                    input_class=ttk.Spinbox,
                    input_var=tk.IntVar(),
                    input_args={"from_": 1, "to": 32, "increment": 1, "width": 5},
                    label_pos="top",
                )
                self.inputs[self.names[i]].grid(
                    row=i, column=0, sticky=tk.NSEW, padx=5, pady=3
                )
            if i > 0:
                self.inputs[self.names[i]] = LabelInput(
                    parent=self,
                    label=self.labels[i],
                    input_class=ttk.Entry,
                    input_var=tk.IntVar(),
                    input_args={"width": 5, "state": "disabled"},
                    label_pos="top",
                )
                self.inputs[self.names[i]].grid(
                    row=i, column=0, sticky=(tk.NSEW), padx=5, pady=3
                )
                self.inputs[self.names[i]].configure(width=5)

    def get_variables(self):
        """This function returns a dictionary of all the variables that are tied to
        each  widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        variables : dict
            The dictionary that holds the variables.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get()
        return variables

    def get_widgets(self):
        """This function returns the dictionary that holds the widgets.

        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        self.inputs : dict
            The dictionary that holds the widgets.
        """
        return self.inputs


class IntensityFrame(ttk.Labelframe):
    """This class is the frame that holds the intensity controls."""

    def __init__(self, cam_view, *args, **kwargs):
        """Initialize the IntensityFrame class.

         Parameters
        ----------
        cam_view : tk.Frame
            The frame that will hold the intensity controls.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        # Init Frame
        text_label = "LUT"
        ttk.Labelframe.__init__(self, cam_view, text=text_label, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: dict: The dictionary that holds the widgets.
        self.inputs = {}

        #: list: The list of LUTs for the image display.
        self.color_labels = [
            "Gray",
            "Gradient",
            "Rainbow",
            "SNR",
        ]

        #: list: The list of maplotlib LUT names.
        self.color_values = [
            "gist_gray",
            "plasma",
            "afmhot",
            "RdBu_r",
        ]

        #: tk.StringVar: The variable that holds the LUT.
        self.color = tk.StringVar()
        for i in range(len(self.color_labels)):
            self.inputs[self.color_labels[i]] = LabelInput(
                parent=self,
                label=self.color_labels[i],
                input_class=ttk.Radiobutton,
                input_var=self.color,
                input_args={"value": self.color_values[i]},
            )
            self.inputs[self.color_labels[i]].grid(
                row=i, column=0, sticky=tk.NSEW, pady=3
            )

        #: tk.BooleanVar: The variable that holds the flip xy flag.
        self.transpose = tk.BooleanVar()

        #: str: The name of the flip xy flag.
        self.trans = "Flip XY"
        self.inputs[self.trans] = LabelInput(
            parent=self,
            label=self.trans,
            input_class=ttk.Checkbutton,
            input_var=self.transpose,
        )
        self.inputs[self.trans].grid(
            row=len(self.color_labels), column=0, sticky=tk.NSEW, pady=3
        )

        #: tk.BooleanVar: The variable that holds the autoscale flag.
        self.autoscale = tk.BooleanVar()

        #: str: The name of the autoscale flag.
        self.auto = "Autoscale"

        #: list: The list of min and max counts.
        self.minmax = ["Min Counts", "Max Counts"]

        #: list: The list of min and max names.
        self.minmax_names = ["Min", "Max"]
        self.inputs[self.auto] = LabelInput(
            parent=self,
            label=self.auto,
            input_class=ttk.Checkbutton,
            input_var=self.autoscale,
        )
        self.inputs[self.auto].grid(
            row=len(self.color_labels) + 1, column=0, sticky=tk.NSEW, pady=3
        )

        # Max and Min Counts
        for i in range(len(self.minmax)):
            self.inputs[self.minmax_names[i]] = LabelInput(
                parent=self,
                label=self.minmax[i],
                input_class=ttk.Spinbox,
                input_var=tk.IntVar(),
                input_args={"from_": 1, "to": 2**16 - 1, "increment": 1, "width": 5},
            )
            self.inputs[self.minmax_names[i]].grid(
                row=i + len(self.color_labels) + 2,
                column=0,
                sticky=tk.NSEW,
                padx=3,
                pady=3,
            )

    def get_variables(self):
        """This function returns a dictionary of all the variables that are tied to
        each  widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        variables : dict
            The dictionary that holds the variables.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get()
        return variables

    def get_widgets(self):
        """This function returns the dictionary that holds the widgets.

        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        self.inputs : dict
            The dictionary that holds the widgets.
        """
        return self.inputs



from pyvcam import pvc
from pyvcam.camera import Camera


def main():
    # Initialize PVCAM and find the first available camera.
    pvc.init_pvcam()
    cam = [cam for cam in Camera.detect_camera()][0]
    cam.open()
    cam.speed_table_index = 0
    for i in range(5):
        frame = cam.get_frame(exp_time=20)
        print("First five pixels of frame: {}, {}, {}, {}, {}".format(*frame[:5]))
    cam.close()
    pvc.uninit_pvcam()

if __name__ == "__main__":
    main()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Imports
import tkinter as tk
from tkinter import ttk, Grid
import logging
from pathlib import Path

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.hovermixin import HoverTkButton
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.view.custom_widgets.validation import ValidatedSpinbox
from navigate.view.custom_widgets.validation import ValidatedEntry


# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class StageControlNotebook(ttk.Notebook):
    """Notebook for stage control tab."""

    def __init__(self, frame_bot_right, *args, **kwargs):
        """Initialize the stage control notebook.

        Parameters
        ----------
        frame_bot_right : tk.Frame
            Frame to put notebook into.
        *args
            Arguments for ttk.Notebook
        **kwargs
            Keyword arguments for ttk.Notebook
        """
        # Init notebook
        ttk.Notebook.__init__(self, frame_bot_right, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        # Putting notebook 3 into bottom right frame
        self.grid(row=0, column=0)

        #: StageControlTab: Stage control tab.
        self.stage_control_tab = StageControlTab(self)

        # Adding tabs to notebook
        self.add(self.stage_control_tab, text="Stage Control", sticky=tk.NSEW)


class GoToFrame(ttk.Frame):
    """GoTo Frame for stage control tab."""

    def __init__(self, stage_control_tab, *args, **kwargs):
        """Initialize the goto frame.

        Parameters
        ----------
        stage_control_tab : StageControlTab
            Stage control tab.
        *args
            Arguments for ttk.Frame
        **kwargs
            Keyword arguments for ttk.Frame
        """
        # Init Frame
        ttk.Frame.__init__(self, stage_control_tab, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)


class StageControlTab(tk.Frame):
    """Stage Control Tab for stage control notebook."""

    def __init__(self, note3, *args, **kwargs):
        """Initialize the stage control tab.

        Parameters
        ----------
        note3 : ttk.Notebook
            Stage control notebook.
        *args
            Arguments for tk.Frame
        **kwargs
            Keyword arguments for tk.Frame
        """
        # Init Frame
        tk.Frame.__init__(self, note3, *args, **kwargs)

        #: int: Index of the stage control tab.
        self.index = 2

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: PositionFrame: Position frame.
        self.position_frame = PositionFrame(self)

        #: XYFrame: XY frame.
        self.xy_frame = XYFrame(self)

        #: OtherAxisFrame: Z frame.
        self.z_frame = OtherAxisFrame(self, "Z")

        #: OtherAxisFrame: Theta frame.
        self.theta_frame = OtherAxisFrame(self, "Theta")

        # OtherAxisFrame: Focus frame.
        self.f_frame = OtherAxisFrame(self, "Focus")

        #: GoToFrame: GoTo frame.
        self.goto_frame = GoToFrame(self)
        self.goto_frame_label = ttk.Label(self.goto_frame, text="Goto Frame")
        self.goto_frame_label.pack()  # For visual mockup purposes

        #: StopFrame: Stop frame.
        self.stop_frame = StopFrame(self, "Stop")

        # Formatting
        tk.Grid.columnconfigure(self.position_frame, "all", weight=1)
        tk.Grid.rowconfigure(self.position_frame, "all", weight=1)
        tk.Grid.columnconfigure(self.xy_frame, "all", weight=1)
        tk.Grid.rowconfigure(self.xy_frame, "all", weight=1)
        tk.Grid.columnconfigure(self.z_frame, "all", weight=1)
        tk.Grid.rowconfigure(self.z_frame, "all", weight=1)
        tk.Grid.columnconfigure(self.theta_frame, "all", weight=1)
        tk.Grid.rowconfigure(self.theta_frame, "all", weight=1)
        tk.Grid.columnconfigure(self.f_frame, "all", weight=1)
        tk.Grid.rowconfigure(self.f_frame, "all", weight=1)
        tk.Grid.columnconfigure(self.goto_frame, "all", weight=1)
        tk.Grid.rowconfigure(self.goto_frame, "all", weight=1)
        tk.Grid.columnconfigure(self.stop_frame, "all", weight=1)
        tk.Grid.rowconfigure(self.stop_frame, "all", weight=1)

        # Griding out frames
        self.position_frame.grid(row=0, column=0, sticky=(tk.NSEW), padx=3, pady=3)
        self.xy_frame.grid(row=0, column=1, sticky=(tk.NSEW), padx=3, pady=3)
        self.z_frame.grid(row=0, column=2, sticky=(tk.NSEW), padx=3, pady=3)
        self.f_frame.grid(row=1, column=0, sticky=(tk.NSEW), padx=3, pady=3)
        self.theta_frame.grid(row=1, column=2, sticky=(tk.NSEW), padx=3, pady=3)
        # self.goto_frame.grid(row=0, column=4, sticky=(tk.NSEW))
        self.stop_frame.grid(row=1, column=1, sticky=(tk.NSEW), padx=3, pady=3)

        # example hover description
        self.xy_frame.up_y_btn.hover.setdescription(
            "Increases the Y value of the stage's position"
        )
        self.xy_frame.down_y_btn.hover.setdescription(
            "Decreases the Y value of the stage's position"
        )
        self.xy_frame.down_x_btn.hover.setdescription(
            "Decreases the X value of the stage's position"
        )
        self.xy_frame.up_x_btn.hover.setdescription(
            "Increases the X value of the stage's position"
        )
        self.position_frame.inputs["y"].widget.hover.setdescription(
            "Y position of the stage"
        )
        self.position_frame.inputs["x"].widget.hover.setdescription(
            "X position of the stage"
        )
        self.position_frame.inputs["z"].widget.hover.setdescription(
            "Z position of the stage"
        )
        self.position_frame.inputs["f"].widget.hover.setdescription("Focus")
        self.stop_frame.joystick_btn.hover.setdescription(
            "Enables/disables joystick mode"
        )

    def get_widgets(self):
        """Get all widgets in the stage control tab.

        This function will return all the input widgets as a dictionary
        The reference name in the dictionary is the same as in the widget list file

        Returns
        -------
        dict
            Dictionary of widgets
        """
        temp = {**self.position_frame.get_widgets()}
        for axis in ["xy", "z", "theta", "f"]:
            temp[axis + "_step"] = getattr(self, axis + "_frame").get_widget()
        return temp

    def get_variables(self):
        """Get all variables in the stage control tab.

        Returns
        -------
        dict
            Dictionary of variables
        """
        temp = self.get_widgets()
        return {k: temp[k].get_variable() for k in temp}

    def get_buttons(self):
        """Get all buttons in the stage control tab.

        this function returns all the buttons in a dictionary

        the reference name is the same as in widget list

        Returns
        -------
        dict
            Dictionary of buttons
        """
        result = {**self.xy_frame.get_buttons()}
        for axis in ["z", "theta", "f"]:
            temp = getattr(self, axis + "_frame").get_buttons()
            result.update({k + "_" + axis + "_btn": temp[k] for k in temp})
        result.update(self.stop_frame.get_buttons())
        return result

    def toggle_button_states(self, joystick_is_on=False, joystick_axes=[]):
        """Enables/disables buttons and entries in the stage control tab,
        according to joystick axes.

        Parameters
        ----------
        joystick_is_on : bool
            'True' indicates that joystick mode is on

            'False' indicates that joystick mode is off

        joystick_axes : ListProxy
            A ListProxy containing the axes controlled by the joystick, if any


        """
        self.xy_frame.toggle_button_states(joystick_is_on, joystick_axes)
        self.z_frame.toggle_button_states(joystick_is_on, joystick_axes)
        self.f_frame.toggle_button_states(joystick_is_on, joystick_axes)
        self.theta_frame.toggle_button_states(joystick_is_on, joystick_axes)
        self.position_frame.toggle_entry_states(joystick_is_on, joystick_axes)
        self.stop_frame.toggle_button_states(joystick_is_on, joystick_axes)

    def force_enable_all_axes(self):
        """Enable all buttons and entries in the stage control tab."""
        self.xy_frame.toggle_button_states(False, ["x", "y"])
        self.z_frame.toggle_button_states(False, ["z"])
        self.f_frame.toggle_button_states(False, ["f"])
        self.theta_frame.toggle_button_states(False, ["theta"])
        self.position_frame.toggle_entry_states(False, ["x", "y", "z", "theta", "f"])
        self.stop_frame.toggle_button_states(False, [])


class OtherAxisFrame(ttk.Labelframe):
    """Frame for the other axis movement buttons.

    This frame is used for the z, theta, and focus axis movement buttons.
    """

    def __init__(self, stage_control_tab, name, *args, **kwargs):
        """Initialize the other axis frame.

        Parameters
        ----------
        stage_control_tab : tk.Frame
            The stage control tab that the other axis frame is in
        name : str
            The name of the axis that the frame is for
        *args : tuple
            Positional arguments for the ttk.Labelframe
        **kwargs : dict
            Keyword arguments for the ttk.Labelframe
        """
        # Init Frame
        label = name
        ttk.Labelframe.__init__(
            self,
            stage_control_tab,
            text=label + " Movement",
            *args,
            **kwargs,
        )

        #: str: Name of the axis.
        self.name = name

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        image_directory = Path(__file__).resolve().parent

        #: tk.PhotoImage: Image for the up button.
        self.up_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greyup.png")
        )

        #: tk.PhotoImage: Image for the down button.
        self.down_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greydown.png")
        )

        #: tk.PhotoImage: Image for the disabled up button.
        self.d_up_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greyup_disabled.png")
        )

        #: tk.PhotoImage: Image for the disabled down button.
        self.d_down_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greydown_disabled.png")
        )

        self.up_image = self.up_image.subsample(2, 2)
        self.down_image = self.down_image.subsample(2, 2)

        #: list: List of images for the normal state.
        self.normal_images = [
            self.up_image,
            self.down_image,
        ]

        self.d_up_image = self.d_up_image.subsample(2, 2)
        self.d_down_image = self.d_down_image.subsample(2, 2)

        self.disabled_images = [
            self.d_up_image,
            self.d_down_image,
        ]

        # Setting up buttons for up, down, zero and increment spinbox

        #: HoverTkButton: Up button.
        self.up_btn = HoverTkButton(self, image=self.up_image, borderwidth=0)

        #: HoverTkButton: Down button.
        self.down_btn = HoverTkButton(self, image=self.down_image, borderwidth=0)

        #: HoverTkButton: Zero button.
        self.zero_btn = HoverTkButton(
            self,
            text="ZERO " + self.name,
        )

        #: LabelInput: Increment spinbox.
        self.increment_box = LabelInput(
            parent=self,
            input_class=ValidatedSpinbox,
            input_var=tk.DoubleVar(),
            input_args={"width": 5},
        )

        #: str: Text for the hover description.
        self.hover_text_ending = ""

        if self.name == "Z":
            self.hover_text_ending = f"the {self.name} \
                value of the stage's position"
        elif self.name == "Theta":
            self.hover_text_ending = "the angle of sample rotation"
        else:
            self.hover_text_ending = f"the {self.name.lower()}"

        # Set the Hover Descriptions
        self.up_btn.hover.setdescription(f"Increases {self.hover_text_ending}")
        self.down_btn.hover.setdescription(f"Decreases {self.hover_text_ending}")

        #: list: List of hover texts for the normal state.
        self.normal_hover_texts = [
            self.up_btn.hover.getdescription(),
            self.down_btn.hover.getdescription(),
        ]

        #: list: List of hover texts for the disabled state.
        self.disabled_hover_texts = [
            "Turn off Joystick Mode to enable",
            "Turn off Joystick Mode to enable",
        ]

        # Griding out buttons
        self.up_btn.grid(row=0, column=0, pady=2)  # UP
        self.down_btn.grid(row=3, column=0, pady=2)  # DOWN
        self.increment_box.grid(row=2, column=0, pady=2)

        # Increment spinbox
        self.increment_box.widget.set_precision(-1)

    def get_widget(self):
        """Returns the frame widget

        Returns
        -------
        tk.Frame
            The frame widget
        """
        return self.increment_box

    def get_buttons(self):
        """Returns the buttons in the frame

        Returns
        -------
        dict
            A dictionary of the buttons
        """
        return {
            "up": self.up_btn,
            "down": self.down_btn,
            "zero": self.zero_btn,
        }

    def toggle_button_states(self, joystick_is_on=False, joystick_axes=[]):
        """Switches the images used as buttons between two states

        Parameters
        ----------
        self : OtherAxisFrame
            The OtherAxisFrame object

        joystick_is_on : bool
            'True' indicates that joystick mode is on

            'False' indicates that joystick mode is off

        joystick_axes : ListProxy
            A ListProxy containing the axes controlled by the joystick, if any
        """

        buttons = [self.up_btn, self.down_btn]
        button_state = "normal"
        image_list = self.normal_images
        hover_list = self.normal_hover_texts

        if joystick_is_on:
            if (self.name.lower() in joystick_axes) or (
                self.name.lower() == "focus" and "f" in joystick_axes
            ):
                button_state = "disabled"
                image_list = self.disabled_images
                hover_list = self.disabled_hover_texts

        for k in range(len(buttons)):
            buttons[k]["state"] = button_state
            buttons[k].config(image=image_list[k])
            buttons[k].hover.setdescription(hover_list[k])


class PositionFrame(ttk.Labelframe):
    """Frame for the stage position entries.

    This frame is used for the x, y, z, theta, and focus position entries.
    """

    def __init__(self, stage_control_tab, *args, **kwargs):
        """Initialize the position frame.

        Parameters
        ----------
        stage_control_tab : tk.Frame
            The stage control tab that the position frame is in
        *args : tuple
            Positional arguments for the ttk.Labelframe
        **kwargs : dict
            Keyword arguments for the ttk.Labelframe
        """

        # Init Frame
        ttk.Labelframe.__init__(
            self, stage_control_tab, text="Stage Positions", *args, **kwargs
        )

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: dict: Dictionary of the label input widgets for the position entries.
        self.inputs = {}

        #: list: List of frames for the position entries.
        entry_names = ["x", "y", "z", "theta", "f"]

        #: list: List of labels for the position entries.
        entry_labels = ["X", "Y", "Z", "\N{Greek Capital Theta Symbol}", "F"]

        #: list: List of frames for the position entries.
        self.frame_back_list = []
        for i in range(len(entry_names)):
            self.inputs[entry_names[i]] = LabelInput(
                parent=self,
                label=entry_labels[i],
                input_class=ValidatedEntry,
                input_var=tk.DoubleVar(),
                input_args={
                    "required": True,
                    "precision": 0.1,
                    "width": 6,
                    "takefocus": False,
                },
            )
            self.frame_back_list.append(
                tk.Frame(self, bg="#f0f0f0", width=60, height=26)
            )
            self.frame_back_list[i].grid(row=i, column=0)
            self.inputs[entry_names[i]].grid(row=i, column=0)
            self.frame_back_list[i].lower()

    def get_widgets(self):
        """Get all widgets in the position frame

        Returns
        -------
        dict
            A dictionary of the label input widgets for the position entries
        """
        return self.inputs

    def get_variables(self):
        """Get all variables in the position frame

        Returns
        -------
        dict
            A dictionary of the variables for the position entries
        """
        variables = {}
        for name in self.inputs:
            variables[name] = self.inputs[name].get_variable()
        return variables

    def toggle_entry_states(self, joystick_is_on=False, joystick_axes=[]):
        """Switches the images used as buttons between two states

        Parameters
        ----------
        joystick_is_on : bool
            'True' indicates that joystick mode is on
            'False' indicates that joystick mode is off
        joystick_axes : ListProxy
            A ListProxy containing the axes controlled by the joystick, if any
        """
        frame_back_counter = 0
        if joystick_is_on:
            entry_state = "disabled"
            frame_back_color = "#ee868a"
        else:
            entry_state = "normal"
            frame_back_color = "#f0f0f0"

        for variable in self.get_variables():
            if variable in joystick_axes:
                self.frame_back_list[frame_back_counter]["bg"] = frame_back_color
                try:
                    self.inputs[f"{variable}"].widget["state"] = entry_state

                except KeyError:
                    pass
            frame_back_counter += 1


class XYFrame(ttk.Labelframe):
    """Frame for the x and y movement buttons.

    This frame is used for the up, down, left, right, zero, and increment buttons.
    """

    def __init__(self, stage_control_tab, *args, **kwargs):
        """Initialize the XY frame.

        Parameters
        ----------
        stage_control_tab : tk.Frame
            The stage control tab that the XY frame is in
        *args : tuple
            Positional arguments for the ttk.Labelframe
        **kwargs : dict
            Keyword arguments for the ttk.Labelframe
        """

        # Init Frame
        ttk.Labelframe.__init__(
            self, stage_control_tab, text="X Y Movement", *args, **kwargs
        )

        # Formatting
        Grid.columnconfigure(self, "all", weight=1)
        Grid.rowconfigure(self, "all", weight=1)

        # Setting up buttons for up, down, left, right, zero and increment spinbox
        s = ttk.Style()
        s.configure("arrow.TButton", font=(None, 20))

        # Path to arrows
        image_directory = Path(__file__).resolve().parent

        #: tk.PhotoImage: Image for the up button.
        self.up_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greyup.png")
        )

        #: tk.PhotoImage: Image for the down button.
        self.down_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greydown.png")
        )

        #: tk.PhotoImage: Image for the left button.
        self.left_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greyleft.png")
        )

        #: tk.PhotoImage: Image for the right button.
        self.right_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greyright.png")
        )

        #: tk.PhotoImage: Image for the disabled up button.
        self.d_up_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greyup_disabled.png")
        )

        #: tk.PhotoImage: Image for the disabled down button.
        self.d_down_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greydown_disabled.png")
        )

        #: tk.PhotoImage: Image for the disabled left button.
        self.d_left_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greyleft_disabled.png")
        )

        #: tk.PhotoImage: Image for the disabled right button.
        self.d_right_image = tk.PhotoImage(
            file=image_directory.joinpath("images", "greyright_disabled.png")
        )

        self.right_image = self.right_image.subsample(2, 2)
        self.left_image = self.left_image.subsample(2, 2)
        self.up_image = self.up_image.subsample(2, 2)
        self.down_image = self.down_image.subsample(2, 2)

        #: list: List of images for the normal state.
        self.normal_images = [
            self.right_image,
            self.left_image,
            self.up_image,
            self.down_image,
        ]

        #: list: List of hover texts for the normal state.
        self.normal_hover_texts = [
            "Decreases the X value of the stage's position",
            "Increases the X value of the stage's position",
            "Increases the Y value of the stage's position",
            "Decreases the Y value of the stage's position",
        ]

        self.d_right_image = self.d_right_image.subsample(2, 2)
        self.d_left_image = self.d_left_image.subsample(2, 2)
        self.d_up_image = self.d_up_image.subsample(2, 2)
        self.d_down_image = self.d_down_image.subsample(2, 2)

        #: list: List of images for the disabled state.
        self.disabled_images = [
            self.d_right_image,
            self.d_left_image,
            self.d_up_image,
            self.d_down_image,
        ]

        #: list: List of hover texts for the disabled state.
        self.disabled_hover_texts = [
            "Turn off Joystick Mode to enable" for i in range(4)
        ]

        #: HoverTkButton: Up button.
        self.up_y_btn = HoverTkButton(
            self,
            image=self.up_image,
            borderwidth=0
        )

        #: HoverTkButton: Down button.
        self.down_y_btn = HoverTkButton(
            self,
            image=self.down_image,
            borderwidth=0
        )

        #: HoverTkButton: Right button.
        self.up_x_btn = HoverTkButton(
            self,
            image=self.right_image,
            borderwidth=0
        )

        #: HoverTkButton: Left button.
        self.down_x_btn = HoverTkButton(
            self,
            image=self.left_image,
            borderwidth=0
        )

        #: LabelInput: Increment spinbox.
        self.increment_box = LabelInput(
            parent=self,
            input_class=ValidatedSpinbox,
            input_var=tk.DoubleVar(),
            input_args={"width": 5},
        )

        #: dict: Dictionary of the buttons for the x and y movement buttons.
        self.button_axes_dict = {
            "x": [self.up_x_btn, self.down_x_btn],
            "y": [self.up_y_btn, self.down_y_btn],
        }

        # Griding out buttons
        self.up_y_btn.grid(
            row=0, column=2, rowspan=2, columnspan=2, padx=2, pady=2
        )  # UP
        self.up_x_btn.grid(
            row=2, column=4, rowspan=2, columnspan=2, padx=2, pady=2
        )  # RIGHT
        self.down_y_btn.grid(
            row=4, column=2, rowspan=2, columnspan=2, padx=2, pady=2
        )  # DOWN
        self.down_x_btn.grid(
            row=2, column=0, rowspan=2, columnspan=2, padx=2, pady=2
        )  # LEFT
        self.increment_box.grid(
            row=3, column=2, rowspan=1, columnspan=2, padx=2, pady=2
        )

        # Increment spinbox
        self.increment_box.widget.set_precision(-1)

    def get_widget(self):
        """Returns the frame widget

        Returns
        -------
        tk.Frame
            The frame widget
        """

        return self.increment_box

    def get_buttons(self):
        """Returns the buttons in the frame

        Returns
        -------
        dict
            A dictionary of the buttons
        """

        names = ["up_x_btn", "down_x_btn", "up_y_btn", "down_y_btn"]
        return {k: getattr(self, k) for k in names}

    def toggle_button_states(self, joystick_is_on=False, joystick_axes=[]):
        """Switches the images used as buttons between two states

        joystick_is_on : bool
            False if buttons are normal, True if buttons are disabled
        joystick_axes : list
            Contains strings representing joystick axes
        """

        for axis in self.button_axes_dict.keys():
            if axis in joystick_axes:
                buttons = self.button_axes_dict[axis]
                axis_ascii = ord(axis)
                if joystick_is_on:
                    button_state = "disabled"
                    image_list = self.disabled_images
                    hover_list = self.disabled_hover_texts
                else:
                    button_state = "normal"
                    image_list = self.normal_images
                    hover_list = self.normal_hover_texts
                for k in range(len(buttons)):
                    buttons[k]["state"] = button_state
                    buttons[k].config(image=image_list[2 * (axis_ascii % 2) + k])
                    buttons[k].hover.setdescription(
                        hover_list[2 * (axis_ascii % 2) + k]
                    )


class StopFrame(ttk.Frame):
    """Frame for the stop button."""

    def __init__(self, stage_control_tab, name, *args, **kwargs):
        """Initialize the stop frame.

        Parameters
        ----------
        stage_control_tab : StageControlTab
            Stage control tab.
        name : str
            Name of the frame.
        *args
            Arguments for ttk.Frame
        **kwargs
            Keyword arguments for ttk.Frame
        """

        # Init Frame
        ttk.Frame.__init__(self, stage_control_tab, *args, **kwargs)

        #: str: Name of the frame.
        self.name = name

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: tk.Button: Stop button.
        self.stop_btn = tk.Button(
            self, bg="red", fg="white", text="STOP", width=20, height=6
        )

        #: HoverTkButton: Joystick button.
        self.joystick_btn = HoverTkButton(
            self, bg="white", fg="black", text="Enable Joystick", width=15, height=2
        )

        # Griding out buttons
        self.stop_btn.grid(row=0, column=0, rowspan=2, pady=2)
        self.joystick_btn.grid(row=2, column=0, rowspan=2, pady=2)

    def get_buttons(self):
        """Returns the buttons in the frame

        Returns
        -------
        dict
            A dictionary of the buttons
        """
        return {"stop": self.stop_btn, "joystick": self.joystick_btn}

    def toggle_button_states(stop_frame, joystick_is_on=False, joystick_axes=[]):
        """Switches the images used as buttons between two states

        Parameters
        ----------
        stop_frame : StopFrame
            The StopFrame object

        joystick_is_on : bool
            'True' indicates that joystick mode is on

            'False' indicates that joystick mode is off

        joystick_axes : ListProxy
            A ListProxy containing the axes controlled by the joystick, if any
        """
        if joystick_is_on:
            stop_frame.joystick_btn.config(text="Disable Joystick")
        else:
            stop_frame.joystick_btn.config(text="Enable Joystick")



import numpy as np
import threading
import time

from pyvcam import pvc
from pyvcam.camera import Camera

NUM_FRAMES = 5
EXPOSE_TIME_ms = 200

def printFrameDetails(frame, fps, frame_count, desc):
    low = np.amin(frame['pixel_data'])
    high = np.amax(frame['pixel_data'])
    average = np.average(frame['pixel_data'])
    print(desc + ': Min:{}\tMax:{}\tAverage:{:.0f}\tFrame Count: {:.0f}'.format(low, high, average, frame_count))

def main():
    # Initialize PVCAM and find the first available camera.
    pvc.init_pvcam()

    cam = [cam for cam in Camera.detect_camera()][0]
    cam.open()

    cam.start_seq(exp_time=EXPOSE_TIME_ms, num_frames=NUM_FRAMES)

    # Wait for 3 frames to be captured. Frames take a slightly longer than expose time, so wait a little extra time
    time.sleep(3 * EXPOSE_TIME_ms * 1e-3 + 0.1)

    cnt = 0
    while cnt < NUM_FRAMES:
        frame, fps, frame_count = cam.poll_frame(oldestFrame=False)
        printFrameDetails(frame, fps, frame_count, 'newest')

        frame, fps, frame_count = cam.poll_frame()
        printFrameDetails(frame, fps, frame_count, 'oldest')

        cnt += 1

    cam.finish()
    cam.close()
    pvc.uninit_pvcam()

    print('\nTotal frames: {}'.format(cnt))

if __name__ == "__main__":
    main()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports

# Third Party Imports
from PIL import Image, ImageDraw, ImageFont
import numpy as np

# Local Imports


def text_array(text: str, offset: tuple = (0, 0)):
    """Create a binary array from a piece of text

    Use Default font to avoid OS related errors.

    Parameters
    ----------
    text : str
        Text to convert
    offset : tuple
        (x,y) offset from upper left of image
    font_size: int
        Size of font in pixels

    Returns
    -------
    np.array
        Binary array of text
    """
    font = ImageFont.load_default()
    bbox = font.getbbox(text)
    im = Image.new(mode="1", size=(bbox[2] + offset[0], bbox[3] + offset[1]))
    ImageDraw.Draw(im).text(xy=offset, text=text, fill=1, font=font)
    return np.array(im, dtype=bool)


def create_arrow_image(xys, image_width=300, image_height=200, direction="right", image=None):
    """Create/Update a Image Object 

    Draw lines and arrows in a Image ojbect

    Parameters
    ----------
    xys : list
        list of points [(x, y),] to draw lines
    image_width : int
        width of image
    image_height: int
        height of image
    direction: str
        arrow directions: "left", "right", "up", "down"
    image: Image/None
        update an exist Image object/create a new Image object

    Returns
    -------
    image:
        Image object
    """
    w, h = image_width, image_height
    if not image:
        image = Image.new("RGBA", (w, h), (0, 0, 0, 0))
    draw = ImageDraw.Draw(image)
    # draw line
    for i in range(len(xys)-1):
        draw.line([xys[i], xys[i+1]], fill="black", width=2)
    
    # draw arrow
    circle_x, circle_y = xys[-1]
    if direction == "right":
        bounding_circle = ((circle_x-10, circle_y), 10)
        rotation = 270
    elif direction == "left":
        bounding_circle = ((circle_x+10, circle_y), 10)
        rotation = 90
    elif direction == "up":
        bounding_circle = ((circle_x, circle_y+10), 10)
        rotation = 0
    elif direction == "down":
        bounding_circle = ((circle_x, circle_y-10), 10)
        rotation = 180
    draw.regular_polygon(bounding_circle, n_sides=3, rotation=rotation, fill="black")
    
    return image


# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard library imports
import os
import inspect
import importlib

# Third-party imports

# Local application imports
from navigate.model.features.auto_tile_scan import CalculateFocusRange  # noqa
from navigate.model.features.autofocus import Autofocus  # noqa
from navigate.model.features.common_features import (
    ChangeResolution,  # noqa
    Snap,  # noqa
    WaitToContinue,  # noqa
    LoopByCount,  # noqa
    PrepareNextChannel,  # noqa
    MoveToNextPositionInMultiPositionTable,  # noqa
    StackPause,  # noqa
    ZStackAcquisition,  # noqa
    FindTissueSimple2D,  # noqa
)
from navigate.model.features.image_writer import ImageWriter  # noqa
from navigate.model.features.restful_features import IlastikSegmentation  # noqa
from navigate.model.features.volume_search import VolumeSearch  # noqa
from navigate.model.features.remove_empty_tiles import (
    DetectTissueInStack,  # noqa
    DetectTissueInStackAndReturn, #noqa
    DetectTissueInStackAndRecord,  # noqa
    RemoveEmptyPositions,  # noqa
)
from navigate.tools.file_functions import load_yaml_file
from navigate.tools.common_functions import load_module_from_file


class SharedList(list):
    """Custom list class with a name attribute for sharing data.

    This class extends the built-in list class to provide an additional attribute
    called 'name' for identifying and sharing lists of data. It is particularly
    useful when sharing data between different parts of a program.

    Notes:
    ------
    - This class is designed to be used when you need to pass and share lists of
      data between different components of a program.
    - The '__name__' attribute can be used to give the shared list a meaningful
      name for easier identification.
    """

    def __init__(self, value, name=None):
        """
        Initialize a SharedList object.

        Parameters:
        ----------
        value : iterable, optional
            Initial values for the list (default is an empty list).
        name : str, optional
            A name to identify the shared list (default is 'shared_list__').
        """
        super().__init__(value)
        if name is None:
            name = "shared_list__"
        self.__name__ = name

    def __str__(self):
        """Return a string representation of the shared list.

        Returns:
        -------
        str
            A string containing the type, name, and current values of the shared
            list.

        Notes:
        ------
        The string representation is in the form of a dictionary with 'type',
        'name', and 'value' keys.
        """

        return str({"type": "shared_list", "name": self.__name__, "value": self})


def convert_str_to_feature_list(content: str):
    """Convert string to a feature list

    Parameters
    ----------
    content : str
        A string value that represents a feature list.

    Returns
    -------
    feature_list : List
        A list: If the string value can be converted to a valid feature list
        None: If can not.
    """

    def convert_args_to_tuple(feature_list):
        """Recursively convert 'args' within a feature list to tuples.

        This function takes a feature list, which is typically used for specifying
        configuration options, and ensures that all 'args' elements within the list
        are converted to tuples. It operates recursively to handle nested feature
        lists.

        Parameters:
        ----------
        feature_list : list
            The feature list to process. It can contain dictionaries with 'args'
            elements that need to be converted to tuples.

        Returns:
        -------
        None

        Notes:
        ------
        - This function operates in-place and does not return a new list.
        - It is useful for ensuring that 'args' elements within a feature list are
          of tuple type, as required by certain functions or processes.
        - The function does not modify elements that are not dictionaries or do not
          have an 'args' key.
        - It handles nested feature lists, converting 'args' in inner lists as well.
        """

        if not feature_list:
            return
        for item in feature_list:
            if type(item) is dict:
                if "args" in item and type(item["args"]) is not tuple:
                    item["args"] = (item["args"],)
            else:
                convert_args_to_tuple(item)

    if content in ["break", '"break"', "'break'"]:
        return "break"
    if content in ["continue", '"continue"', "'continue'"]:
        return "continue"
    try:
        exec_result = {}
        exec(f"result={content}", globals(), exec_result)
        if type(exec_result["result"]) is not list:
            print("Please make sure the feature list is a list!")
            return None
        # 'args' should be tuple
        convert_args_to_tuple(exec_result["result"])
        return exec_result["result"]
    except Exception as e:
        print("Can't build this feature list!", e)
        return None


def convert_feature_list_to_str(feature_list):
    """Convert a feature list to its string representation.

    This function takes a feature list, which is typically used for specifying
    configuration options, and converts it to its string representation. The
    resulting string can be useful for saving, displaying, or transmitting
    feature lists.

    Parameters:
    ----------
    feature_list : list
        A valid feature list.

    Returns:
    -------
    result : str
        The string representation of the feature list.

    Notes:
    ------
    - The function recursively processes the input feature list, converting
      dictionaries, tuples, and lists to their corresponding string
      representations.
    - Within the resulting string, dictionaries are represented with keys
      "name" and "args" (if present), where "name" is the name of the feature
      function and "args" contains the arguments as a tuple.
    - Tuples and lists are represented using parentheses and square brackets,
      respectively.
    - This function is intended for converting feature lists to a format that
      can be easily saved to a file, transmitted over a network, or displayed
      to users.

    Example:
    --------
    Given the following feature list:

    ```
    [
        {"name": func1, "args": (arg1, arg2)},
        {"name": func2},
        [
            {"name": func3, "args": (arg3,)},
            {"name": func4}
        ]
    ]
    ```

    The function would return the string representation:

    ```
    "[{'name': 'func1', 'args': ('arg1', 'arg2')}, {'name': 'func2'},
    [{'name': 'func3', 'args': ('arg3',)}, {'name': 'func4'}]]"
    ```
    """
    if feature_list == "break" or feature_list == "continue":
        return f'"{feature_list}"'

    result = "["

    def f(feature_list):
        """Recursively convert a feature list to its string representation."""
        if not feature_list:
            return
        nonlocal result
        for item in feature_list:
            if type(item) is dict:
                result += "{" + f'"name": {item["name"].__name__},'
                if "args" in item:
                    result += '"args": ('
                    for temp in item["args"]:
                        if temp is None:
                            result += "None,"
                        elif callable(temp):
                            result += f'"{temp.__name__}",'
                        elif type(temp) is str:
                            result += f'"{temp}",'
                        else:
                            result += f"{temp},"
                    result += "),"
                if "true" in item:
                    if type(item["true"]) == str:
                        result += f'"true": "{item["true"]}",'
                    else:
                        result += '"true":' + convert_feature_list_to_str(item["true"]) + ','
                if "false" in item:
                    if type(item["false"]) == str:
                        result += f'"false": "{item["false"]}",'
                    else:
                        result += '"false":' + convert_feature_list_to_str(item["false"]) + ','
                result += "},"
            elif type(item) is tuple:
                result += "("
                f(item)
                result += "),"
            elif type(item) is list:
                result += "["
                f(item)
                result += "],"

    f(feature_list)
    result += "]"
    return result


def load_dynamic_parameter_functions(
    feature_list: list, feature_parameter_setting_path: str
):
    """Load dynamic parameter functions into a feature list.

    This function takes a feature list and a path to the parameter setting files
    and dynamically updates the feature list with parameter values specified in
    the setting files. It searches for parameter values in the setting files
    and replaces corresponding arguments in the feature list with the loaded
    values. The function is designed to support dynamic configuration of
    feature functions with external parameter files.

    Parameters:
    -----------
    feature_list : list
        A list of feature configurations, typically used for specifying
        configuration options. It may contain dictionaries with "name" and "args"
        keys, where "name" is the feature function and "args" is a tuple of
        arguments. Arguments that match parameter names specified in the setting
        files will be replaced with loaded values.

    feature_parameter_setting_path : str
        The path to the directory containing parameter setting files. These files
        should be named based on the feature function names (e.g., function_name.yml).

    Notes:
    ------
    - The function iterates through the feature list, inspecting each feature
      dictionary for parameter values specified in the setting files.
    - Parameters that match argument names in the feature list are replaced with
      the loaded values from the setting files.
    - The function supports loading parameter values from YAML files and
      dynamically updating the feature list.

    Example:
    --------
    Given a feature list and a parameter setting file:
    - feature_list:
      ```
      [
          {"name": func1, "args": (arg1, arg2)},
          {"name": func2},
          [
              {"name": func3, "args": (arg3,)},
              {"name": func4}
          ]
      ]
      ```
    - parameter setting file (func1.yml):
      ```
      arg1: value1
      arg2: value2
      ```

    The function would update the feature list as follows:
    ```
    [
        {"name": func1, "args": ("value1", "value2")},
        {"name": func2},
        [
            {"name": func3, "args": (arg3,)},
            {"name": func4}
        ]
    ]
    ```
    """
    for item in feature_list:
        if type(item) is dict:
            if "args" in item:
                feature = item["name"]
                config_path = (
                    f"{feature_parameter_setting_path}/{feature.__name__}.yml"
                )
                parameter_config = None
                if os.path.exists(config_path):
                    parameter_config = load_yaml_file(config_path)
                if parameter_config:
                    args = list(item["args"])
                    spec = inspect.getfullargspec(feature)
                    for parameter in parameter_config:
                        idx = spec.args[2:].index(parameter)
                        if idx >= len(args):
                            continue
                        if args[idx] is None:
                            args[idx] = "None"
                        if args[idx] not in parameter_config[parameter]:
                            ref_lib = None
                        else:
                            ref_lib = parameter_config[parameter][args[idx]]
                        if ref_lib is None or ref_lib == "None":
                            args[idx] = None
                        elif os.path.exists(ref_lib):
                            module = load_module_from_file(args[idx], ref_lib)
                            args[idx] = getattr(module, args[idx]) if module else None
                        else:
                            module = importlib.import_module(ref_lib)
                            args[idx] = getattr(module, args[idx]) if module else None
                    item["args"] = tuple(args)
        elif item == "break" or item == "continue":
            continue
        else:
            load_dynamic_parameter_functions(item, feature_parameter_setting_path)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
from tkinter import ttk
import logging

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.popup import PopUp
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.tools.common_functions import build_ref_name

p = __name__.split(".")[1]
logger = logging.getLogger(p)


class MicroscopeSettingPopupWindow:
    """Popup window with waveform parameters for galvos, remote focusing, etc."""

    def __init__(self, root, microscope_info, *args, **kwargs):
        """Initialize the popup window

        Parameters
        ----------
        root : tkinter.Tk
            Root window
        microscope_info : dict
            Dictionary of microscope information
        *args :
            Variable length argument list
        **kwargs :
            Arbitrary keyword arguments
        """

        # Creating popup window with this name and size/placement, PopUp is a
        #: PopUp: PopUp window class
        self.popup = PopUp(
            root, "Configure Microscopes", "+320+180", top=False, transient=False
        )

        # Storing the content frame of the popup, this will be the parent of
        # the widgets
        content_frame = self.popup.get_frame()
        content_frame.columnconfigure(0, pad=5)
        content_frame.columnconfigure(1, pad=5)
        content_frame.rowconfigure(0, pad=5)
        content_frame.rowconfigure(1, pad=5)
        content_frame.rowconfigure(2, pad=5)

        # Formatting
        tk.Grid.columnconfigure(content_frame, "all", weight=1)
        tk.Grid.rowconfigure(content_frame, "all", weight=1)

        """Creating the widgets for the popup"""
        #: dict: Dictionary of input widgets
        self.inputs = {}
        #: dict: Dictionary of buttons
        self.buttons = {}

        # Frames for widgets
        label_frame = ttk.Frame(content_frame, padding=(0, 0, 0, 0))
        #: ttk.Frame: Frame for microscope information
        self.microscopes_frame = ttk.Frame(content_frame, padding=(0, 0, 0, 0))
        button_frame = ttk.Frame(content_frame, padding=(0, 0, 0, 0))

        # Griding Frames
        label_frame.grid(row=0, column=0, sticky=tk.NSEW)
        self.microscopes_frame.grid(row=0, column=1, sticky=tk.NSEW)
        button_frame.grid(row=1, column=1, sticky=tk.SE)

        #: list: List of labels
        self.labels = ["Microscope Name"]
        for microscope_name in microscope_info:
            for k in microscope_info[microscope_name].keys():
                device = " ".join(map(lambda v: v.capitalize(), k.split("_")))
                if device not in self.labels:
                    self.labels.append(device)
        self.labels.append("Setting")

        for i, name in enumerate(self.labels):
            row = ttk.Label(label_frame, text=name, padding=(2, 0, 0, 0))
            if i == 0:
                row.grid(row=i, column=0, padx=(0, 5), pady=(5, 30))
            else:
                row.grid(row=i, column=0, padx=(0, 5), pady=(0, 22))

        self.list_microscope_info(microscope_info)

        # button
        self.buttons["confirm"] = ttk.Button(button_frame, text="Confirm")
        self.buttons["confirm"].grid(
            row=0, column=0, sticky=tk.SE, padx=(5, 0), pady=(0, 5)
        )
        self.buttons["cancel"] = ttk.Button(button_frame, text="Cancel")
        self.buttons["cancel"].grid(
            row=0, column=1, sticky=tk.SE, padx=(5, 0), pady=(0, 5)
        )

    def list_microscope_info(self, microscope_info):
        """List the microscope information in the popup window

        Parameters
        ----------
        microscope_info : dict
            Dictionary of microscope information
        """
        c = 0
        for microscope_name in microscope_info.keys():
            frame = ttk.Frame(self.microscopes_frame, padding=(0, 0, 0, 0))
            frame.grid(row=0, column=c, sticky=tk.NSEW)
            c += 1
            m_name = ttk.Label(frame, text=microscope_name, padding=(2, 0, 0, 0))
            m_name.grid(row=0, column=0, sticky=tk.NSEW, padx=(0, 5), pady=(5, 16))
            for i in range(1, len(self.labels) - 1):
                device_ref_name = build_ref_name(
                    "_", *self.labels[i].lower().split(" ")
                )
                entry = LabelInput(
                    parent=frame,
                    input_class=ttk.Entry,
                    input_var=tk.StringVar(),
                    label_args={"padding": (0, 0, 5, 20)},
                )
                entry.set(microscope_info[microscope_name].get(device_ref_name, ""))
                entry.widget["state"] = "disabled"
                entry.grid(row=i, column=0, sticky=tk.NSEW, padx=(2, 5), pady=(2, 0))
                self.inputs[f"{microscope_name} {device_ref_name}"] = entry
            # usage
            combo = LabelInput(
                parent=frame,
                input_class=ttk.Combobox,
                input_var=tk.StringVar(),
                label_args={"padding": (0, 0, 5, 20)},
            )
            combo.grid(
                row=len(self.labels), column=0, sticky=tk.SE, padx=(2, 5), pady=(2, 0)
            )
            self.inputs[f"{microscope_name}"] = combo

    # Getters
    def get_variables(self):
        """This function returns a dictionary of all the variables tied to each widget.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        dict
            Dictionary of variables

        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get_variable()
        return variables

    def get_widgets(self):
        """This function returns the dictionary that holds the input widgets.

        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        dict
            Dictionary of widgets

        """
        return self.inputs

    def get_buttons(self):
        """This function returns the dictionary that holds the buttons.

        The key is the button name, value is the button.

        Returns
        -------
        dict
            Dictionary of buttons

        """
        return self.buttons



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
from tkinter import Menu
import logging

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


#  Menubar class
class Menubar(Menu):
    """Menu bar class for GUI"""

    def __init__(self, window, *args, **kwargs):
        """Initialize the menu bar class.


        Parameters
        ----------
        window : tkinter.Tk
            Parent window for menu bar
        *args : list
            List of arguments
        **kwargs : dict
            Dictionary of arguments

        """

        #  Init Menu with parent
        Menu.__init__(self, window, *args, **kwargs)

        #: str: Operating system of computer
        self.opsystem = window.tk.call("tk", "windowingsystem")

        #  Prevents menu from tearing off bar
        window.option_add("*tearOff", False)

        #  Linking menu to option of parent to this menu class
        window["menu"] = self

        #: Menu: File Menu
        self.menu_file = Menu(self)
        self.add_cascade(menu=self.menu_file, label="File")

        #: Menu: Microscope Configuration Menu
        self.menu_resolution = Menu(self)
        self.add_cascade(menu=self.menu_resolution, label="Microscope Configuration")

        #: Menu: Stage Control Menu
        self.menu_multi_positions = Menu(self)
        self.add_cascade(menu=self.menu_multi_positions, label="Stage Control")

        #: Menu: Autofocus Menu
        self.menu_autofocus = Menu(self)
        self.add_cascade(menu=self.menu_autofocus, label="Autofocus")

        #: Menu: Features Menu
        self.menu_features = Menu(self)
        self.add_cascade(menu=self.menu_features, label="Features")

        #: Menu: Plugins Menu
        self.menu_plugins = Menu(self)
        self.add_cascade(menu=self.menu_plugins, label="Plugins")

        #: Menu: Window Menu
        self.menu_window = Menu(self)
        self.add_cascade(menu=self.menu_window, label="Window")



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Imports
import logging
import time
from serial import SerialException

# Local Imports
from navigate.model.devices.stages.stage_base import StageBase
from navigate.model.devices.APIs.sutter.MP285 import MP285

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


def build_MP285_connection(com_port, baud_rate, timeout=0.25):
    """Build Sutter Stage Serial Port connection

    Parameters
    ----------
    com_port : str
        COM Port for the SutterStage.
    baud_rate : int
        Baud Rate for the SutterStage.
    timeout : float
        Duration of time for timeout to be triggered (s)

    Returns
    -------
    MP285
        MP285 SutterStage.
    """
    try:
        mp285_stage = MP285(com_port, baud_rate, timeout)
        mp285_stage.connect_to_serial()
        return mp285_stage
    except SerialException as e:
        logger.debug(f"Sutter MP-285 - Could not establish Serial Port Connection: {e}")
        raise UserWarning(
            "Could not communicate with Sutter MP-285 via COMPORT", com_port
        )


class SutterStage(StageBase):
    """SutterStage Class for MP-285."""

    def __init__(self, microscope_name, device_connection, configuration, device_id=0):
        """Initialize the SutterStage.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : MP285
            MP285 stage.
        configuration : dict
            Configuration dictionary for the SutterStage.

        Raises
        ------
        UserWarning
            Error while connecting to the SutterStage.
        UserWarning
            Error while setting resolution and velocity.
        UserWarning
            Error while setting absolute operation mode.
        """
        super().__init__(microscope_name, device_connection, configuration, device_id)

        # Device Connection
        if device_connection is None:
            logger.error("The MP285 stage is unavailable!")
            raise UserWarning("The MP285 stage is unavailable!")

        #: object: MP285 stage.
        self.stage = device_connection

        #: bool: Wait until stage has finished moving before returning.
        self.stage.wait_until_done = True

        # Default mapping from self.axes to corresponding MP285 axis labelling
        axes_mapping = {"x": "x", "y": "y", "z": "z"}
        if not self.axes_mapping:
            #: dict: Dictionary of stage axes and their corresponding hardware axes.
            self.axes_mapping = {
                axis: axes_mapping[axis] for axis in self.axes if axis in axes_mapping
            }

        #: dict: Dictionary of hardware axes and their corresponding stage axes.
        self.device_axes = dict(map(lambda v: (v[1], v[0]), self.axes_mapping.items()))

        # Default Operating Parameters
        #: str: Resolution of the stage.
        self.resolution = "low"  # "high"

        #: int: Speed of the stage.
        self.speed = 3000  # 1300  # in units microns/s.

        #: float: Position of the stage along the x-axis.
        #: float: Position of the stage along the y-axis.
        #: float: Position of the stage along the z-axis.
        self.stage_x_pos, self.stage_y_pos, self.stage_z_pos = None, None, None

        # Set the resolution and velocity of the stage
        try:
            self.stage.set_resolution_and_velocity(
                resolution=self.resolution, speed=self.speed
            )
        except Exception as e:
            logger.debug(f"Sutter MP-285 - Error setting resolution and velocity: {e}")
            raise UserWarning("Sutter MP-285 - Error setting resolution and velocity")

        # Set the operating mode of the stage.
        try:
            self.stage.set_absolute_mode()
        except Exception as e:
            logger.debug(f"Sutter MP-285 - Error setting absolute operation mode: {e}")
            raise UserWarning("Sutter MP-285 - Error setting absolute operation mode")

        self.report_position()

    def __del__(self):
        """Delete SutterStage Serial Port.



        Raises
        ------
        UserWarning
            Error while closing the SutterStage Serial Port.
        """
        self.close()

    def report_position(self):
        """Reports the position for all axes, and creates a position dictionary.

        Positions from the MP-285 are converted to microns.

        Returns
        -------
        position : dict
            Dictionary containing the position of all axes
        """
        position = {}
        try:
            (
                self.stage_x_pos,
                self.stage_y_pos,
                self.stage_z_pos,
            ) = self.stage.get_current_position()
            for axis, hardware_axis in self.axes_mapping.items():
                hardware_position = getattr(self, f"stage_{hardware_axis}_pos")
                self.__setattr__(f"{axis}_pos", hardware_position)

            position = self.get_position_dict()
            logger.debug(f"MP-285 - Position: {position}")
        except SerialException as e:
            print("MP-285: Failed to report position: {e}")
            logger.debug(f"MP-285 - Error: {e}")
            time.sleep(0.01)

        return position

    def move_axis_absolute(self, axis, abs_pos, wait_until_done=False):
        """Implement movement logic along a single axis.

        Parameters
        ----------
        axis : str
            An axis. For example, 'x', 'y', 'z', 'f', 'theta'.
        abs_pos : float
            Absolute position value
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        bool
            Was the move successful?
        """
        move_dictionary = {f"{axis}_abs": abs_pos}
        return self.move_absolute(move_dictionary, wait_until_done)

    def move_absolute(self, move_dictionary, wait_until_done=True):
        """Move stage along a single axis.

        Parameters
        ----------
        move_dictionary : dict
            A dictionary of values required for movement. Includes 'x_abs', 'x_min',
            etc. for one or more axes. Expects values in micrometers, except for theta,
            which is in degrees.
        wait_until_done : bool
            Wait until the stage has finished moving before returning.

        Returns
        -------
        bool
            Was the move successful?
        """
        pos_dict = self.verify_abs_position(move_dictionary)
        if not pos_dict:
            return False

        # rely on cached positions
        # if len(pos_dict.keys()) < 3:
        #     self.report_position()
        self.stage.wait_until_done = wait_until_done
        move_stage = {}
        for axis in pos_dict:
            if (
                abs(
                    getattr(self, f"stage_{self.axes_mapping[axis]}_pos")
                    - pos_dict[axis]
                )
                < 0.02
            ):
                move_stage[axis] = False
            else:
                move_stage[axis] = True
                setattr(self, f"stage_{self.axes_mapping[axis]}_pos", pos_dict[axis])

        move_stage = any(move_stage.values())
        if move_stage is True:
            try:
                self.stage.move_to_specified_position(
                    x_pos=self.stage_x_pos,
                    y_pos=self.stage_y_pos,
                    z_pos=self.stage_z_pos,
                )
            except SerialException as e:
                logger.debug(f"MP285: move_axis_absolute failed - {e}")
                # make sure the cached positions are the "same" as device
                self.report_position()
                return False

        return True

    def stop(self):
        """Stop all stage movement abruptly."""
        try:
            self.stage.interrupt_move()
        except SerialException as error:
            logger.exception(f"MP-285 - Stage stop failed: {error}")

    def close(self):
        """Close the stage."""

        try:
            self.stop()
            self.stage.close()
            logger.debug("MP-285 stage connection closed")
        except (AttributeError, BaseException) as e:
            print("Error while closing the MP-285 stage connection", e)
            logger.debug("Error while disconnecting the MP-285 stage", e)
